import time


class LimitUsageManager(object):
    def __init__(self, limits, items):
        self.limits = limits
        self.max_time_limit = max(self.limits.keys())
        self.usage_times = {item: [] for item in items}

    def is_active(self, item):
        if item not in self.usage_times.keys():
            raise Exception("Wrong item")

        for time_interval, limit in self.limits.items():
            usage_times = self.usage_times[item]

            if len(usage_times) < limit:
                continue

            last_limit_usage = usage_times[-limit]
            if time.time() - last_limit_usage < time_interval:
                return False

        return True

    def _clean_old_usages(self, item):
        usage_times = self.usage_times[item]
        old_usages_count = 0
        for usage_time in usage_times:
            if usage_time > time.time() - self.max_time_limit:
                break

            old_usages_count += 1

        self.usage_times[item] = usage_times[old_usages_count:]

    def add_usage(self, item):
        self.usage_times[item].append(time.time())
        self._clean_old_usages(item)


if __name__ == "__main__":
    limit_usage_manager = LimitUsageManager({2: 2, 3: 3}, ["first", "second", "third"])

    assert limit_usage_manager.is_active("first")
    assert limit_usage_manager.is_active("second")
    assert limit_usage_manager.is_active("third")

    limit_usage_manager.add_usage("first")
    limit_usage_manager.add_usage("first")
    limit_usage_manager.add_usage("first")
    limit_usage_manager.add_usage("second")
    limit_usage_manager.add_usage("second")
    limit_usage_manager.add_usage("third")

    assert not limit_usage_manager.is_active("first")
    assert not limit_usage_manager.is_active("second")
    assert limit_usage_manager.is_active("third")

    time.sleep(2)

    assert not limit_usage_manager.is_active("first")
    assert limit_usage_manager.is_active("second")
    assert limit_usage_manager.is_active("third")

    time.sleep(1)

    assert limit_usage_manager.is_active("first")
    assert limit_usage_manager.is_active("second")
    assert limit_usage_manager.is_active("third")

    print("Test LimitUsageManager OK")



import time
import logging
import zmq

from utils_ak.str import b, u
from utils_ak.coder import JsonCoder

CONTEXT = zmq.Context()

logger = logging.getLogger(__name__)


def endpoint(host, port):
    return f"tcp://{host}:{port}"


def split_endpoint(endpoint):
    protocol, host, port = endpoint.split(":")
    return protocol, host[2:], port


def cast_connection(address, conn_type):
    protocol, host, port = split_endpoint(address)
    if conn_type == "auto":
        conn_type = "bind" if "*" in address else "connect"
    elif conn_type == "bind":
        address = f"tcp://*:{port}"
    elif conn_type == "connect":
        if host == "*":
            raise Exception(f"Cannot connect to specified address: {address}")
    return address, conn_type


def connect_socket(socket, address, conn_type):
    address, conn_type = cast_connection(address, conn_type)
    # "tcp://*:5563"
    if conn_type == "bind":
        socket.bind(address)
    elif conn_type == "connect":
        socket.connect(address)
    else:
        raise Exception(f"Connection type {conn_type} not recognized")


class Publisher(object):
    def __init__(
        self,
        address,
        sndhwm=0,
        init_timeout=0.1,
        context=CONTEXT,
        conn_type="auto",
        json_coder=JsonCoder(),
    ):
        self.context = context
        self.socket = context.socket(zmq.PUB)
        self.address = address
        self.json_coder = json_coder

        if sndhwm is not None:
            # self.socket.SNDHWM = sndhwm
            self.socket.setsockopt(zmq.SNDHWM, sndhwm)
        connect_socket(self.socket, address, conn_type)
        time.sleep(init_timeout)

    def term(self):
        self.socket.close()

    def publish(self, topic, msg):
        # todo: fix msg stuff!!!
        self.socket.send_multipart([b(topic), b(msg)])

    def publish_json(self, topic, msg):
        self.socket.send_multipart([b(topic), b(self.json_coder.encode(msg))])


class Subscriber(object):
    def __init__(self, address, rcvhwm=None, context=CONTEXT, conn_type="auto"):
        # Prepare our context and publisher
        self.context = context
        self.socket = context.socket(zmq.SUB)
        self.address = address

        if rcvhwm is not None:
            self.socket.setsockopt(zmq.RCVHWM, rcvhwm)

        connect_socket(self.socket, address, conn_type)

        self.topics = []

    def subscribe(self, topic, timeout=0.01):
        topic = b(topic)

        if topic in self.topics:
            logger.warning(f"Already subscribed for topic {u(topic)}")
            return

        self.socket.setsockopt(zmq.SUBSCRIBE, topic)
        time.sleep(timeout)

    def receive(self):
        multipart_msg = self.socket.recv_multipart()
        topic, contents = multipart_msg[-2:]
        return topic, contents

    def term(self):
        self.socket.close()


class Forwarder(object):
    """ For multi pub-sub communication. """

    def __init__(self, endpoint_in, endpoint_out, context=CONTEXT):
        self.endpoint_in = endpoint_in
        self.endpoint_out = endpoint_out
        self.context = context

    def run(self):
        try:
            # Socket facing clients
            logging.info(
                f"Forwarder. Frontend (IN): {self.endpoint_in} Backend (OUT): {self.endpoint_out}"
            )
            self.frontend_socket = self.context.socket(zmq.SUB)
            connect_socket(self.frontend_socket, self.endpoint_in, "bind")
            self.frontend_socket.setsockopt(zmq.SUBSCRIBE, b(""))
            self.backend_socket = self.context.socket(zmq.PUB)
            connect_socket(self.backend_socket, self.endpoint_out, "bind")
            zmq.device(zmq.FORWARDER, self.frontend_socket, self.backend_socket)
        except Exception as e:
            logger.error("Exception occured. Bringing down forwarder device")
        finally:
            self.frontend_socket.close()
            self.backend_socket.close()


class Streamer(object):
    """ For multi push-pull communication. """

    def __init__(self, endpoint_in, endpoint_out, context=CONTEXT):
        self.endpoint_in = endpoint_in
        self.endpoint_out = endpoint_out
        self.context = context

    def run(self):
        try:
            # Socket facing clients
            logging.info(
                f"Streamer. Frontend (IN): {self.endpoint_in} Backend (OUT): {self.endpoint_out}"
            )

            self.frontend_socket = self.context.socket(zmq.PULL)
            connect_socket(self.frontend_socket, self.endpoint_in, "bind")

            self.backend_socket = self.context.socket(zmq.PUSH)
            connect_socket(self.backend_socket, self.endpoint_out, "bind")
            zmq.device(zmq.STREAMER, self.frontend_socket, self.backend_socket)
        except Exception as e:
            logger.error("Exception occured. Bringing down streamer device")
        finally:
            self.frontend_socket.close()
            self.backend_socket.close()



import decimal
import struct
import six

import datetime as datetime_module

STRUCT_DATE = struct.Struct("!HBB")
STRUCT_DATETIME = struct.Struct("!q")
STRUCT_DECIMAL_LENGTH = struct.Struct("!H")
STRUCT_TIME = struct.Struct("!3BL")


def pack_date(obj):
    return STRUCT_DATE.pack(obj.year, obj.month, obj.day)


def unpack_date(data):
    return datetime_module.date(*STRUCT_DATE.unpack(data))


def pack_datetime(obj):
    if obj.tzinfo is not None:
        raise TypeError("Cannot encode time zone-aware date-times to MessagePack")
    seconds = (obj - datetime_module.datetime(1970, 1, 1)).total_seconds()
    microseconds = int(seconds * 1000000.0)
    return STRUCT_DATETIME.pack(microseconds)


def unpack_datetime(data):
    microseconds = STRUCT_DATETIME.unpack(data)[0]
    return datetime_module.datetime.utcfromtimestamp(microseconds / 1000000.0)


def pack_time(obj):
    return STRUCT_TIME.pack(obj.hour, obj.minute, obj.second, obj.microsecond)


def unpack_time(data):
    return datetime_module.time(*STRUCT_TIME.unpack(data))


def pack_decimal(obj):
    obj_str = six.text_type(obj)[:65535].encode("utf-8")
    obj_len = len(obj_str)
    obj_encoder = struct.Struct(f"!H{obj_len}s")
    return obj_encoder.pack(obj_len, obj_str)


def unpack_decimal(data):
    obj_len = STRUCT_DECIMAL_LENGTH.unpack(data[:2])[0]
    obj_decoder = struct.Struct(f"!{obj_len}s")
    return decimal.Decimal(obj_decoder.unpack(data[2:])[0].decode("utf-8"))



class Worker:
    def __init__(self, id, payload):
        self.id = id
        self.payload = payload

    def run(self):
        raise NotImplementedError



def run_worker(worker_cls, worker_config):
    worker = worker_cls(worker_config["worker_id"], worker_config["payload"])
    worker.run()



import asyncio
import websockets
import time
from icecream import ic

from utils_ak.jsonrpc.clients.websockets_client import WebSocketsClient

from utils_ak.id_generator import *

loop = asyncio.get_event_loop()


async def main():
    async with websockets.connect("ws://localhost:5000") as ws:
        client = WebSocketsClient(ws, IncrementalIDGenerator())

        asyncio.ensure_future(client.start_receiving_loop())

        for i in range(5):
            response = await client.execute("ping")
            ic(response)
            time.sleep(3)


asyncio.ensure_future(main())
loop.run_forever()



import time
import multiprocessing

from loguru import logger

from utils_ak.loguru import configure_loguru_stdout
from utils_ak.simple_microservice import run_listener_async
from utils_ak.job_orchestrator.worker.sample_worker.sample_worker import SampleWorker
from utils_ak.job_orchestrator.monitor import Monitor
from utils_ak.job_orchestrator.tests.config.config import config


def run_monitor():
    configure_loguru_stdout("DEBUG")
    logger.info("Running monitor...")
    monitor = Monitor(config.TRANSPORT)
    monitor.microservice.run()


def run_worker():
    configure_loguru_stdout("DEBUG")
    logger.info("Running SampleWorker instance...")
    worker = SampleWorker(
        "WorkerId",
        {"type": "batch", "message_broker": config.TRANSPORT},
    )
    worker.run()


def run():
    configure_loguru_stdout("DEBUG")
    run_listener_async("monitor_out", message_broker=config.TRANSPORT)
    time.sleep(1)

    multiprocessing.Process(target=run_monitor).start()
    time.sleep(3)
    multiprocessing.Process(target=run_worker).start()


if __name__ == "__main__":
    run()



from utils_ak.fluid_flow import *


def test_flow_container_1():
    container1 = Container("Input", value=100, max_pressures=[None, 50])
    container2 = Container("Output")

    pipe_connect(container1, container2)

    flow = FluidFlow(container1, verbose=True)
    run_flow(flow)


def test_flow_container_2():
    container1 = Container(
        "Input", value=100, max_pressures=[None, 50], limits=[None, 30]
    )
    container2 = Container("Output")

    pipe_connect(container1, container2)

    flow = FluidFlow(container1, verbose=True)
    run_flow(flow)


def test_flow_container_3():
    container1 = Container(
        "Input", value=100, max_pressures=[None, 50], limits=[None, 30]
    )
    container2 = Container("Output", max_pressures=[5, None], limits=[20, None])

    pipe_connect(container1, container2)

    flow = FluidFlow(container1, verbose=True)
    run_flow(flow)


if __name__ == "__main__":
    test_flow_container_1()
    # test_flow_container_2()
    # test_flow_container_3()



from .casts import *
from .proxy_guard import ProxyGuard
from .proxy_bag import ProxyBag
from .availability_proxy_guard import AvailabilityProxyGuard
from .time_limit_proxy_guard import TimeLimitProxyGuard



from typing import Any, Optional, Iterator
from apply_defaults import apply_self
from jsonrpcclient.clients.websockets_client import (
    WebSocketsClient as _WebSocketsClient,
)
from jsonrpcclient.response import Response
from jsonrpcclient.requests import Request
from jsonrpcclient.parse import parse

from utils_ak.coder import JsonCoder

import time
import asyncio
from icecream import ic


class WebSocketsClient(_WebSocketsClient):
    """
    A wrapper for jsonrpcclient.websockets_client.WebSocketsClient with proper id matching.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.responses = {}  # {<id>: response}
        self.coder = JsonCoder()

    async def send_message(
        self, request: str, response_expected: bool, **kwargs: Any
    ) -> Response:
        """make response_expected=False for every call"""
        return await super().send_message(request, response_expected=False, **kwargs)

    @apply_self
    async def request(
        self,
        method_name: str,
        *args: Any,
        trim_log_values: bool = False,
        validate_against_schema: bool = True,
        id_generator: Optional[Iterator] = None,
        **kwargs: Any
    ) -> Response:
        request = Request(method_name, id_generator=id_generator, *args, **kwargs)

        id = str(request["id"])

        await self.send(
            request,
            trim_log_values=trim_log_values,
            validate_against_schema=validate_against_schema,
        )

        while True:
            if id in self.responses:
                return self.responses.pop(id)
            await asyncio.sleep(0.001)

    def stop_receiving_loop(self):
        self.responses = None

    async def start_receiving_loop(self):
        while True:
            if self.responses is None:
                return

            try:
                response_text = await asyncio.wait_for(self.socket.recv(), timeout=0.1)
            except asyncio.TimeoutError:
                await asyncio.sleep(0)
                continue
            data = self.coder.decode(response_text)
            id = str(data["id"])

            if self.responses is None:
                return

            self.responses[id] = data
            await asyncio.sleep(0)



import pandas as pd
from loguru import logger


def cast_prop_values(parent, child, key):
    pv = None if not parent else parent[key]
    v = child.relative_props.get(key)
    return pv, v


def relative_acc(parent, child, key, default=None, formatter=None):
    if key in child.relative_props:
        res = child.relative_props[key]
    else:
        if callable(default):
            default = default()
        res = default

    if formatter:
        res = formatter(res)
    return res


def cumsum_acc(parent, child, key, default=None, formatter=None):
    res = relative_acc(parent, child, key, default=default, formatter=formatter)
    if parent:
        res += parent[key]
    return res


class DynamicProps:
    def __init__(
        self,
        props=None,
        formatters=None,
        accumulators=None,
        required_keys=None,
        cache_keys=None,
    ):
        self.accumulators = accumulators or {}
        self.required_keys = required_keys or []
        self.cache_keys = cache_keys or []
        self.formatters = formatters or {}

        self.relative_props = self._format_props(props or {})

        self.parent = None
        self.children = []

        self.cache = {}

    def _format_props(self, props):
        props = dict(props)
        for k, fmt in self.formatters.items():
            if k in props:
                props[k] = fmt(k, props[k])
        return props

    def reset_cache(self, recursive=False):
        self.cache = {}
        if recursive:
            for child in self.children:
                child.reset_cache(recursive=True)

    @staticmethod
    def default_accumulator(parent, child, key):
        pv, v = cast_prop_values(parent, child, key)
        return v if v is not None else pv

    def update(self, **props):
        props = self._format_props(props)
        self.relative_props.update(**props)
        self.reset_cache(recursive=True)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self
        child.reset_cache(recursive=True)

    def remove_child(self, child):
        self.children.remove(child)
        child.parent = None
        child.reset_cache(recursive=True)

    def __getitem__(self, item):
        if item in self.cache:
            return self.cache[item]

        accumulator = self.accumulators.get(item, self.default_accumulator)
        res = accumulator(self.parent, self, item)

        if item in self.cache_keys:
            self.cache[item] = res
        return res

    def get(self, item, default=None):
        res = self[item]
        if res is None:
            res = default
        return res

    def keys(self):
        parent_keys = [] if not self.parent else self.parent.keys()
        res = (
            parent_keys
            + list(self.accumulators.keys())
            + list(self.relative_props.keys())
            + self.required_keys
        )
        return list(set(res))

    def all(self):
        return {key: self[key] for key in self.keys()}


def test_dynamic_props():
    ACCUMULATORS = {"<cumsum>": cumsum_acc, "<relative>": relative_acc}

    def gen_props(props=None):
        return DynamicProps(
            props=props,
            accumulators=ACCUMULATORS,
            required_keys=["<cumsum>", "<relative>", "<other>"],
            cache_keys=["<cumsum>"],
        )

    root = gen_props({"<cumsum>": 1, "<relative>": 5, "<other>": 1})
    child1 = gen_props({"<cumsum>": 2})
    child2 = gen_props({"<cumsum>": 3})
    root.add_child(child1)
    child1.add_child(child2)

    def print_values():
        values = []
        for i, node in enumerate([root, child1, child2]):
            values.append(
                [
                    node[key]
                    for key in [
                        "<cumsum>",
                        "<relative>",
                        "<other>",
                        "<non-existent_key>",
                    ]
                ]
            )
        print(
            pd.DataFrame(
                values,
                columns=["<cumsum>", "<relative>", "<other>", "<non-existent_key>"],
            )
        )

    print_values()
    print_values()

    root.remove_child(child1)
    root.add_child(child1)

    print_values()

    for node in [root, child1, child2]:
        print(node.keys(), node.all())


if __name__ == "__main__":
    test_dynamic_props()



import functools
from loguru import logger
from utils_ak.loguru import configure_loguru_stdout


def log_function(*, entry=True, exit=True, level="DEBUG"):
    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            logger_ = logger.opt(depth=1)
            if entry:
                logger_.log(
                    level,
                    "Entering function",
                    function_name=name,
                    args=args,
                    kwargs=kwargs,
                )
            result = func(*args, **kwargs)
            if exit:
                logger_.log(
                    level, "Exiting function", function_name=name, result=result
                )
            return result

        return wrapped

    return wrapper


def test():
    configure_loguru_stdout("DEBUG")

    @log_function()
    def f(*args, **kwargs):
        logger.info("Inside the function")
        return 1

    f("a", b=1)


if __name__ == "__main__":
    test()



# from confluent_kafka import Producer, Consumer, TopicPartition
import uuid


# https://github.com/edenhill/librdkafka/wiki/Manually-setting-the-consumer-start-offset


class KafkaAdapter:
    """ Allows to read. """

    def __init__(self, bootstrap_servers="localhost:9092", auto_offset_reset="largest"):
        self.kafka_topics = []
        self.state = None
        self.consumer = Consumer(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": str(uuid.uuid4()),
                "default.topic.config": {"auto.offset.reset": auto_offset_reset},
                "enable.auto.commit": False,
                "enable.partition.eof": False,
            }
        )

        self.topic = None
        self.topic_partition = None

    def subscribe(self, topic, offset=0, partition=0):
        self.topic = topic
        self.topic_partition = TopicPartition(topic, partition, offset)
        self.consumer.assign([self.topic_partition])
        # self.seek()

    def seek(self):
        self.consumer.seek(self.topic_partition)

    def receive(self, number):
        return self.consumer.consume(number, timeout=1)

    def has_messages(self):
        smallest, largest = self.get_edge_offsets()
        # TODO: check for by-one blunders
        return self.consumer.position([self.topic_partition])[0].offset <= largest

    def get_edge_offsets(self):
        smallest, largest = self._get_watermarks()
        return smallest, largest

    # Get the interval in which offsets reside
    def _get_watermarks(self):
        return self.consumer.get_watermark_offsets(self.topic_partition)


if __name__ == "__main__":
    adapter = KafkaAdapter()
    adapter.subscribe("benchmark1", 99999)

    i = 0
    while True:
        msg = adapter.receive(1)
        if not msg:
            continue
        msg = msg[0]
        print(msg.offset())
        i += 1
        if i == 10:
            break

    # message = a.consumer.consume(1)[0]
    # print(message.error(), message.value(), message.topic())



import time
import os
from utils_ak.loguru import logger, configure_loguru_stdout
from utils_ak.deployment.io import *


def _test_controller(controller_cls):
    configure_loguru_stdout("DEBUG")

    deployment_fn = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "examples/hello_world/deployment.yml",
    )
    deployment = read_deployment(deployment_fn)
    ctrl = controller_cls()
    logger.info("Starting")
    ctrl.start(deployment)
    time.sleep(3)
    logger.info("Logs")
    ctrl.log(deployment["id"])
    logger.info("Stopping")
    ctrl.stop(deployment["id"])



from utils_ak.id_generator.id_generator import IDGenerator
import uuid


class IncrementalIDGenerator(IDGenerator):
    def __init__(self, start_from=0):
        self.cur_id = start_from

    def gen_id(self):
        res = self.cur_id
        self.cur_id += 1
        return res



from utils_ak.dag import *

from utils_ak.clock import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import *
from utils_ak.fluid_flow.actors.container import Container
from utils_ak.fluid_flow.calculations import *
from utils_ak.iteration import SimpleIterator

from functools import wraps


def switch(f):
    @wraps(f)
    def inner(self, *args, **kwargs):
        pipe_switch(self, self.io_containers["in"], "in")
        pipe_switch(self, self.io_containers["out"], "out")

        res = f(self, *args, **kwargs)

        pipe_switch(self, self.io_containers["in"], "in")
        pipe_switch(self, self.io_containers["out"], "out")
        return res

    return inner


class Sequence(Actor, PipeMixin):
    def __init__(self, name, containers):
        super().__init__(name)
        assert len(containers) >= 2
        self.containers = containers
        self.io_containers = {"in": containers[0], "out": containers[-1]}

        self.nodes = [self.containers[0]]
        for c1, c2 in SimpleIterator(self.containers).iter_sequences(2):
            pipe = pipe_connect(c1, c2)
            self.nodes.append(pipe)
            self.nodes.append(c2)

    def is_limit_reached(self, orient):
        return self.io_containers[orient].is_limit_reached(orient)

    def inner_actors(self):
        return self.nodes

    @switch
    def update_value(self, ts):
        for node in self.containers:
            node.update_value(ts)

    @switch
    def update_pressure(self, ts):
        for node in self.containers:
            node.update_pressure(ts)

    @switch
    def update_speed(self, ts):
        for node in self.nodes:
            node.update_speed(ts)

    @switch
    def update_triggers(self, ts):
        for node in self.containers:
            node.update_triggers(ts)

    def __str__(self):
        return f"Sequence: {self.name}"

    def stats(self):
        return {node.name: node.stats() for node in self.nodes}

    def display_stats(self):
        return [node.display_stats() for node in self.containers]

    def active_periods(self, orient="in"):
        return self.io_containers[orient].active_periods(orient=orient)



import pandas as pd
import numpy as np

import os
import base64

from io import BytesIO

from utils_ak.os import *
from utils_ak.builtin import iter_get, remove_neighbor_duplicates, delistify
from utils_ak.iteration import *

pd.set_option("display.max_colwidth", None)


def display_with_image_thumbnails(df, shape=None):
    from IPython.display import HTML, display
    from PIL import Image

    def _get_thumbnail(path, shape=None):
        if shape is None:
            shape = (150, 150)
        i = Image.open(path)
        i.thumbnail(shape, Image.LANCZOS)
        return i

    def _image_base64(im, shape):
        if isinstance(im, str):
            im = _get_thumbnail(im, shape)
        with BytesIO() as buffer:
            im.save(buffer, "jpeg")
            return base64.b64encode(buffer.getvalue()).decode()

    def _image_formatter(im, shape):
        if type(im) != str:
            return str(im)
        possible_path = im
        if not os.path.exists(possible_path):
            return im

        ext = os.path.splitext(possible_path)[-1]

        if ext not in [".png", ".jpg"]:
            return im

        return '<img src="granular_storage:image/jpeg;base64,{}">'.format(
            _image_base64(im, shape)
        )

    columns = df.columns
    formatters = {col: lambda x: _image_formatter(x, shape) for col in columns}
    # displaying PIL.Image objects embedded in dataframe
    display(HTML(df.to_html(formatters=formatters, escape=False)))


def where(cond, x, y):
    """np.where analogue for dataframes with the same signature. Return elements, either from x or y, depending on condition.

    Parameters
    ----------
    cond: boolean condition
    cond_value: value to return when cond is True
    other: value to return when cond is False

    Returns
    -------
    pd.DataFrame
        np.where wrapped pandas dataframe
    """
    if isinstance(x, pd.DataFrame):
        df = x
    elif isinstance(y, pd.DataFrame):
        df = y
    elif isinstance(cond, pd.DataFrame):
        df = cond
    else:
        raise Exception("Either cond, x, y must be a dataframe")
    return pd.DataFrame(np.where(cond, x, y), index=df.index, columns=df.columns)


def series_equal(s1, s2, dtype="float", eps=1e-8):
    if dtype == "float":
        return s1.sub(s2).abs() < eps
    else:
        return s1 == s2


def find_row(df, row):
    """Find last occurence of row in df
    :df: pd.DataFrame
    :row: pd.DataFrame with unit length
    :return: int. Row number
    """
    if len(row) != 1:
        raise Exception("Row should be one line dataframe")

    if not np.all(df.columns == row.columns):
        raise Exception("Dataframe and row have different columns")

    # filter by index
    # time is considered discrete - no problem on rounding here
    df.index.name = "index"
    df = df.reset_index()
    tmp = df[df["index"] == row.index[0]].copy()
    tmp.pop("index")

    if len(tmp) == 0:
        return None

    # filter by values
    for col in tmp.columns:
        # todo: make better
        dtype = "float" if "float" in str(df.dtypes[0]) else None

        row_series = pd.Series(index=tmp[col].index)
        row_series[:] = row[col].iloc[0]

        tmp[col] = series_equal(tmp[col], row_series, dtype=dtype)
    """ 0    False
        1     True
        2    False
        3    False
        dtype: bool"""
    row_equal = tmp.all(axis=1)
    tmp = tmp[row_equal]

    if len(tmp) == 0:
        return None

    return tmp.index[-1]


def merge_by_columns(dfs):
    res_df = dfs[0].copy()
    for df in dfs[1:]:
        for key in df.columns:
            res_df[key] = df[key]
    return res_df


def merge(dfs, by, keep="last", sort_index=True):
    """
    :param dfs: list(`pd.DataFrame`)
    :param by: name of column or list of columns names. 'all' for all columns. 'columns' for merge_by_columns method
    :param keep: 'last' or 'first'

    :return:
    """
    if not by:
        return pd.concat(dfs, axis=0)

    if isinstance(by, str):
        by = [by]

    by_index = False

    if isinstance(by, list) and "index" in by:
        by.remove("index")
        by_index = True
        by = delistify(by)

    if by == "columns":
        assert not by_index, "Merging by columns only"
        return merge_by_columns(dfs)

    df = pd.concat(dfs, axis=0)

    masks = []
    if by == "all":
        masks.append(df)
    elif by:
        masks.append(df[by])

    if by_index:
        masks.append(pd.Series(df.index, index=df.index))
    mask = pd.concat(masks, axis=1)
    df = df[~mask.duplicated(keep=keep)]

    if sort_index:
        df = df.sort_index()
    return df


def find_gaps(index, gap):
    index = pd.Series(index, index=index)
    index_diff = index - index.shift(1)
    index_diff = index_diff[index_diff > gap]
    return index_diff


# todo: search for existing solutions
def pd_read(fn, index_column=None, **kwargs):
    ext = os.path.splitext(fn)[-1]
    if ".zip" in ext:
        ext = os.path.splitext(fn[:-4])[-1]
    if ext == ".msgpack":
        from mbf_pandas_msgpack import read_msgpack

        df = read_msgpack(fn)
    else:
        df = getattr(pd, f"read_{ext[1:]}")(fn, **kwargs)
    if index_column:
        df = df.set_index(index_column)
    return df


def pd_write(df, fn, index_column=None, **kwargs):
    """NOTE: index is dropped when writing pandas dataframe."""
    if index_column:
        prev_index_column = df.index.name or "index"
        df = df.reset_index()
        df[index_column] = df.pop(prev_index_column)
    ext = os.path.splitext(fn)[-1]
    if ".zip" in ext:
        ext = os.path.splitext(fn[:-4])[-1]
        kwargs["compression"] = "zip"

    kwargs["index"] = False

    if ext == ".hdf" and "key" not in kwargs:
        kwargs["key"] = "key"

    tmp_fn = fn + ".tmp"

    if ext == ".msgpack":
        from mbf_pandas_msgpack import to_msgpack

        res = to_msgpack(tmp_fn, df)
    else:
        res = getattr(df, f"to_{ext[1:]}")(tmp_fn, **kwargs)
    if os.path.exists(fn):
        remove_path(fn)
    rename_path(tmp_fn, fn)
    return res


def mark_consecutive_groups(df, key, groups_key):
    cur_value = None
    cur_i = 0

    values = []
    for i, row in df.iterrows():
        if row[key] != cur_value:
            cur_i += 1
            cur_value = row[key]
        values.append(cur_i)
    df[groups_key] = values


def df_to_tree(df, recursive=True, delistify=False):
    df = df.copy()
    if len(df.columns) == 1:
        res = list(set(df[df.columns[0]].tolist()))
        if delistify and len(res) == 1:
            res = res[0]
        return res

    res = {}

    for value, grp in df.groupby(df.columns[0]):
        grp.pop(df.columns[0])
        if recursive:
            res[value] = df_to_tree(grp, recursive=True, delistify=delistify)
        else:
            res[value] = grp
    return res


def df_to_ordered_tree(df, column=None, recursive=True, prune_last=True):
    df = df.copy()
    column = column or df.columns[0]
    assert column in df.columns

    if column != df.columns[0]:
        other_columns = list(df.columns)
        other_columns.remove(column)
        df = df[[column] + other_columns]

    if len(df.columns) == 1:
        res = df[df.columns[0]].tolist()
        if prune_last:
            res = remove_neighbor_duplicates(res)
        return res

    res = []
    pre_res = []

    mark_consecutive_groups(df, column, "__key")
    for value, grp in df.groupby([column, "__key"]):
        grp.pop(column)
        grp.pop("__key")
        if recursive:
            child = df_to_ordered_tree(grp, recursive=True, prune_last=prune_last)
        else:
            child = grp
        pre_res.append((value[0], value[1], child))

    pre_res = list(
        sorted(
            pre_res,
            key=lambda obj: df[(df[column] == obj[0]) & (df["__key"] == obj[1])].iloc[
                0
            ]["__key"],
        )
    )
    return [(x[0], x[2]) for x in pre_res]


def crop_invalid_edges(df, side="all"):
    if side in ["prefix", "all"]:
        first_idx = df.first_valid_index()
    else:
        first_idx = None

    if side in ["suffix", "all"]:
        last_idx = df.last_valid_index()
    else:
        last_idx = None
    return df.loc[first_idx:last_idx]


# todo: better naming
def split_into_sum_groups(df, values, column, group_column):
    df = df.copy()
    assert abs(sum(values) - df[column].sum()) < 1e-5

    values_iterator = SimpleIterator(values)
    rows_iterator = iter(iter_pairs(list(df.iterrows()), "any_suffix"))

    cur_needed = next(values_iterator)
    cur_row_pair, next_row_pair = next(rows_iterator)
    cur_available = cur_row_pair[-1][column]

    res = []
    i = 0
    while True:
        cur_value = min(cur_needed, cur_available)
        cur_needed -= cur_value
        cur_available -= cur_value

        row = pd.Series(cur_row_pair[-1])
        row[column] = cur_value
        row[group_column] = i
        res.append(dict(row))

        if cur_needed < 1e-5:
            if cur_available < 1e-5 and next_row_pair is None:
                # finished
                break
            else:
                # go to next value
                cur_needed = next(values_iterator)
                i += 1

        if cur_available < 1e-5:
            # go to next row
            cur_row_pair, next_row_pair = next(rows_iterator)
            cur_available = cur_row_pair[-1][column]

    df = pd.DataFrame(res)
    df[group_column] = df[group_column].astype(int)
    return df


def loc_non_inclusive(df, slice):
    df = df.loc[slice]
    mask = df.index.where(df.index != slice.stop)
    mask = pd.Series(mask, index=df.index)
    mask = mask.fillna(method="bfill")
    mask = ~mask.isnull()

    if not mask.any():
        return pd.DataFrame()

    df = df[mask]
    return df


def test_loc_non_inclusive():
    df = pd.DataFrame(dict(A=range(4)), index=[1, 2, 2, 3])
    print(loc_non_inclusive(df, slice(None, 1, None)))
    print(loc_non_inclusive(df, slice(None, 2, None)))
    print(loc_non_inclusive(df, slice(None, 3, None)))
    print(loc_non_inclusive(df, slice(None, 4, None)))


def test_merge():
    df1 = pd.DataFrame.from_dict({"a": [1, 2, 3], "b": [4, 5, 6], "c": [1, 1, 1]})
    df1 = df1.set_index("c")
    df2 = pd.DataFrame.from_dict({"a": [3, 4, 5], "b": [7, 8, 9]})

    print(df1)
    print("-----")
    print(df2)

    print("----------")

    dfs = [df1, df2]

    print("Merge by a")
    print(merge(dfs, by="a"))
    print("Merge by index and a")
    print(merge(dfs, by=["index", "a"]))
    print("Merge by index")
    print(merge(dfs, by="index"))
    print("Merge by index, keep first")
    print(merge(dfs, by="index", keep="first"))
    print("Merge by all and index, keep first")
    print(merge(dfs, by=["all", "index"], keep="first"))


def test_tree():
    df = pd.DataFrame(
        [
            ["A", "a", "a1"],
            ["A", "a", "a2"],
            ["A", "a", "a1"],
            ["B", "b", "b1"],
            ["B", "b", "b2"],
            ["A", "c", "c1"],
            ["A", "c", "c2"],
            ["A", "c", "c2"],
        ],
        columns=["col1", "col2", "col3"],
    )
    print(df)
    print(df_to_tree(df))
    print(df_to_ordered_tree(df))
    print(df_to_ordered_tree(df, prune_last=False))


def test_read_write():
    from datetime import datetime, timedelta

    df = pd.DataFrame.from_dict(
        {"a": [datetime(2020, 1, 1)] * 3, "b": [4, 5, 6], "c": [1, 1, 1]}
    )

    for fn in ["tmp.parquet", "tmp.csv", "tmp.msgpack"]:
        pd_write(df, fn)
        print(pd_read(fn))
        remove_path(fn)

    pd_write(df, "tmp.hdf")
    print(pd_read("tmp.hdf"))
    remove_path("tmp.hdf")


def test_crop_invalid_edges():
    df = pd.DataFrame(
        [
            [np.nan, np.nan, np.nan],
            ["A", np.nan, "a2"],
            ["A", "c", "c2"],
            ["A", "c", "c2"],
            [np.nan, np.nan, np.nan],
        ],
        columns=["col1", "col2", "col3"],
    )

    print(df)
    print(crop_invalid_edges(df, "all"))
    print(crop_invalid_edges(df, "prefix"))
    print(crop_invalid_edges(df, "suffix"))


def test_split_into_sum_groups():
    df = pd.DataFrame.from_dict({"a": [1.0, 2.0, 3.0], "b": ["<1>", "<2>", "<3>"]})
    print(split_into_sum_groups(df, [1.0, 1.5, 1.5, 2.0], "a", "group"))


if __name__ == "__main__":
    # test_merge()
    # test_tree()
    # test_read_write()
    # test_crop_invalid_edges()
    # test_split_into_sum_groups()
    test_loc_non_inclusive()



from confluent_kafka import Producer, Consumer
import uuid
from datetime import datetime, timedelta
from utils_ak.builtin import update_dic
from utils_ak.time import *
from copy import deepcopy
import time
from utils_ak.kafka.kafka_client import KafkaClient


def test():
    cli = KafkaClient(consumer_config={"auto_offset_reset": "smallest"})

    start_dt = datetime.now()
    for i in range(10):
        cli.publish("collection", str(i).encode("utf-8"))
        time.sleep(0.1)
    end_dt = datetime.now()

    cli.flush()
    time.sleep(1)
    cli.subscribe(
        "collection",
        # start_offset=689,
        start_dt=start_dt + (end_dt - start_dt) / 2,
    )

    i = 0
    while True:
        msg = cli.poll()
        if not msg:
            continue
        print(msg)
        print(i)
        i += 1


if __name__ == "__main__":
    test()



from utils_ak.networking.proxy.availability_proxy_guard import AvailabilityProxyGuard
from utils_ak.networking.proxy.limit_usage_manager import LimitUsageManager


class TimeLimitProxyGuard(AvailabilityProxyGuard):
    def __init__(self, proxy_config, *args, proxy_limits=None, **kwargs):
        super().__init__(proxy_config, *args, **kwargs)

        if proxy_limits is None:
            raise Exception("This proxy guard can`t work without limit")

        proxies = proxy_config.keys()
        self.limit_usage_manager = LimitUsageManager(proxy_limits, proxies)

    @property
    def available_counter(self):
        counter = super().available_counter

        limit_exceeded_proxies = []
        for proxy in counter.keys():
            if not self.limit_usage_manager.is_active(proxy):
                limit_exceeded_proxies.append(proxy)

        for proxy in limit_exceeded_proxies:
            del counter[proxy]

        return counter

    def on_acquire_proxy_callback(self, proxy):
        self.limit_usage_manager.add_usage(proxy)


if __name__ == "__main__":
    import time

    config = {"http://localhos:80": 10, "http://192.168.0.1:80": 10}
    # seconds: usage_limit
    limits = {1: 1, 4: 2}
    proxy_guard = TimeLimitProxyGuard(
        config, check_proxy_availability=False, proxy_limits=limits
    )

    time_started = time.time()

    assert len(proxy_guard.available_counter.items()) == 2
    proxy_guard.acquire_proxy()
    proxy_guard.acquire_proxy()
    assert len(proxy_guard.available_counter.items()) == 0

    time.sleep(1)
    assert len(proxy_guard.available_counter.items()) == 2
    proxy_guard.acquire_proxy()
    assert len(proxy_guard.available_counter.items()) == 1

    time.sleep(1)
    assert len(proxy_guard.available_counter.items()) == 1
    proxy_guard.acquire_proxy()
    assert len(proxy_guard.available_counter.items()) == 0

    time.sleep(1)
    assert len(proxy_guard.available_counter.items()) == 0

    time.sleep(1)
    assert len(proxy_guard.available_counter.items()) == 2

    # this assert validate that we not sleep when use acquire_proxy
    assert time.time() - time_started < 4.1

    print("Test TimeLimitProxyGuard is OK")



""" Dictionary with limited size. """

from utils_ak.dict import LastUsedOrderedDict


class LimitedDict(LastUsedOrderedDict):
    def __init__(self, max_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_size = max_size

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

        if self.max_size is not None and len(self) > self.max_size:
            while len(self) > self.max_size:
                self.popitem(last=False)


def test():
    d = LimitedDict(max_size=2)
    d["1"] = 1
    d["2"] = 2
    print(d.pick("1"))
    d["3"] = 3
    print(d)


if __name__ == "__main__":
    test()



from utils_ak.deployment.io import *


def test_read_deployment_file():
    deployment = read_deployment("../examples/hello_world/deployment.yml")
    print(deployment)


if __name__ == "__main__":
    test_read_deployment_file()



import os
import time
import fire
import notifiers

from loguru import logger
from datetime import datetime, timedelta

from utils_ak.deployment.examples.hello_world.config import config


def main(name=None, run_forever=False):
    notifier = notifiers.get_notifier("gmail")

    name = name or os.environ.get("NAME") or "World"
    logger.info(f"Running deployment, {name}")

    for i in range(5):
        notifier.notify(
            message=f"Running deployment, {name}: {datetime.now()}",
            subject="utils_ak/deployment/examples/hello_world",
            to=config.EMAIL_USER,
            username=config.EMAIL_USER,
            password=config.EMAIL_PSWD,
        )
        time.sleep(2)

    if run_forever:
        while True:
            time.sleep(2)


if __name__ == "__main__":
    fire.Fire(main)



from utils_ak.properties import *
from utils_ak.architecture import delistify
from collections import defaultdict


class Block:
    def __init__(
        self,
        block_class=None,
        default_block_class="block",
        props_formatters=None,
        props_accumulators=None,
        props_required_keys=None,
        props_cache_keys=None,
        **props,
    ):
        block_class = block_class or default_block_class
        props["cls"] = block_class

        self.parent = None
        self.children = []
        self.children_by_cls = defaultdict(list)

        self.props = DynamicProps(
            props=props,
            formatters=props_formatters,
            accumulators=props_accumulators,
            required_keys=props_required_keys,
            cache_keys=props_cache_keys,
        )

    def __getitem__(self, item):
        res = self.get(item)
        if not res:
            if isinstance(res, str):
                raise KeyError(item)
            elif isinstance(res, int):
                return IndexError(item)
            else:
                raise Exception(f"Not found: {item}")
        return res

    def get(self, item):
        if isinstance(item, str):
            res = self.children_by_cls[item]
            # res = [b for b in self.children if b.props["cls"] == item]
        elif isinstance(item, int):
            res = self.children[item]
        elif isinstance(item, slice):
            # Get the start, stop, and step from the slice
            res = [self[ii] for ii in range(*item.indices(len(self)))]
        else:
            raise TypeError("Item type not supported")
        return delistify(res)

    def __str__(self):
        res = f'{self.props["cls"]}\n'

        for child in self.children:
            for line in str(child).split("\n"):
                if not line:
                    continue
                res += "  " + line + "\n"
        return res

    def __repr__(self):
        return str(self)

    def query_match(self, query):
        for k, v in query.items():
            if callable(v):
                if not v(self.props[k]):
                    return False
            else:
                if self.props[k] != v:
                    return False
        return True

    def iter(self, **query):
        if self.query_match(query):
            yield self

        for child in self.children:
            for b in child.iter(**query):
                yield b

    def add_child(self, block):
        block.parent = self
        self.children.append(block)
        self.children_by_cls[block.props["cls"]].append(block)
        self.props.add_child(block.props)
        return block

    def remove_child(self, block):
        block.parent = None
        self.children.remove(block)
        self.children_by_cls[block.props["cls"]].remove(block)
        self.props.remove_child(block.props)
        return block


def test_block():
    def cast_block(block_class, **kwargs):
        return Block(
            block_class,
            props_accumulators={
                "t": lambda parent, child, key: cumsum_acc(
                    parent, child, key, default=0, formatter=int
                )
            },
            **kwargs,
        )

    a = cast_block("a", t=5)
    b = cast_block("b")
    c = cast_block("c", t=3)
    a.add_child(b)
    b.add_child(c)

    print(a)
    print(b)
    print(c)
    print(a.props["t"])
    print(b.props["t"])
    print(c.props["t"])

    print()
    print(a["b"]["c"])

    print("Test query")
    for b in a.iter(cls=c):
        print(b)


if __name__ == "__main__":
    test_block()



import os
from utils_ak.coder import *
from utils_ak.dict import *


def create_deployment(
    container_name, deployment_id, payload, image=None, python_main=None
):
    deployment_template = cast_dict_or_list(
        os.path.join(os.path.dirname(__file__), "deployment.yml.template")
    )

    kwargs = {
        "container_name": container_name,
        "deployment_id": str(deployment_id),
        "payload": payload,
        "image": image or "",
        "python_main": python_main or "",
    }
    deployment = fill_template(deployment_template, **kwargs)
    return deployment



class ReturnValuesSink:
    """
    I'm a sink that allows to collect return values from multiple runnables.
    I'm thread-safe unless you call the same pipeline multiple times.
    """

    def __init__(self):
        super().__init__()
        self.results = {}

    def _id(self, callable):
        return id(callable)

    def store(self, mark, result):
        self.results[mark] = result

    def get_pipeline(
        self, mark
    ):  # I give the pipeline. Run it like pipeline("value") to store the value
        def collector(result):
            self.store(mark, result)

        return collector

    def get_results(self):
        return self.results


if __name__ == "__main__":

    def go():
        return "kek"

    sink = ReturnValuesSink()
    pipeline = sink.get_pipeline(
        "for_example_i_store_the_go_return_value_as_this_string"
    )

    pipeline(go())

    print(sink.get_results())



class DAGNode:
    def __init__(self):
        self.parents = []
        self.children = []

        self._iteration_state = {}
        self.reset_iteration_state()

    def reset_iteration_state(self, orient="down", with_children=False):
        self._iteration_state = {"is_processed": False}  # iteration state variable

        if with_children:
            for child in self.oriented_children(orient):
                child.reset_iteration_state(orient, with_children=True)

    def leaves(self, orient="down"):
        res = []
        for node in self.iterate_as_tree(orient):
            if not node.oriented_children(orient):
                res.append(node)
        return res

    def top(self, orient="down"):
        leaves = self.leaves(orient)
        assert len(leaves) == 1, "Top node not found"
        if len(leaves) == 1:
            return leaves[0]

    def root(self, orient="down"):
        cur_node = self

        while True:
            if not cur_node.oriented_children(orient):
                return cur_node
            else:
                cur_node = cur_node.oriented_children(orient)[0]

    def oriented_parents(self, orient="down"):
        return {"down": self.parents, "up": self.children}[orient]

    def oriented_children(self, orient="upwards"):
        return {"down": self.children, "up": self.parents}[orient]

    def iterate(self, orient="down"):
        assert not self.oriented_parents(orient), "Can iterate only from root node"
        self.reset_iteration_state(orient, with_children=True)

        cur_node = self

        while True:
            unprocessed_parents = [
                node
                for node in cur_node.oriented_parents(orient)
                if not node._iteration_state["is_processed"]
            ]
            if unprocessed_parents:
                cur_node = unprocessed_parents[0]
                continue

            yield cur_node
            cur_node._iteration_state["is_processed"] = True

            unprocessed_children = [
                node
                for node in cur_node.oriented_children(orient)
                if not node._iteration_state["is_processed"]
            ]

            if unprocessed_children:
                cur_node = unprocessed_children[0]
                continue
            else:
                break

    @staticmethod
    def _iter_node_as_tree_recursive(cur_node, orient="down"):
        if cur_node._iteration_state["is_processed"]:
            return
        else:
            yield cur_node
            cur_node._iteration_state["is_processed"] = True

        for child in cur_node.oriented_children(orient):
            for node in DAGNode._iter_node_as_tree_recursive(child, orient):
                yield node

    def iterate_as_tree(self, orient="down"):
        assert not self.oriented_parents(orient), "Can iterate only from root node"
        self.reset_iteration_state(orient, with_children=True)

        for node in self._iter_node_as_tree_recursive(self):
            yield node


def connect(parent, child, safe=True):
    if safe and (not parent or not child):
        return
    parent.children.append(child)
    child.parents.append(parent)


def disconnect(parent, child, safe=True):
    if safe and (not parent or not child):
        return
    parent.children.remove(child)
    child.parents.remove(parent)


def test_dag_iteration():
    class NamedNode(DAGNode):
        def __init__(self, name):
            super().__init__()
            self.name = name

        def __str__(self):
            return self.name

        def __repr__(self):
            return self.name

    root_up = NamedNode("root_up")
    node1 = NamedNode("1")
    node2 = NamedNode("2")
    root_down = NamedNode("root_down")

    connect(root_up, node1)
    connect(root_up, node2)
    connect(node1, root_down)
    connect(node2, root_down)

    print("Testing iteration")
    for orient in ["down", "up"]:
        print("Processing orientation", orient)
        for node in [root_up, node1, node2, root_down]:
            print("Processing node", node)
            if orient == "down" and node == root_up:
                for iter_node in node.iterate(orient):
                    print(iter_node)
            elif orient == "up" and node == root_down:
                for iter_node in node.iterate(orient):
                    print(iter_node)
            else:
                try:
                    list(node.iterate(orient))
                except AssertionError as e:
                    print("AssertionError", e)

    print("Testing root")
    for orient in ["down", "up"]:
        print("Processing orientation", orient)
        for node in [root_up, node1, node2, root_down]:
            print(node.root(orient))

    print("Testing Leaves, top, iterate as tree")
    root_up = NamedNode("root_up")
    node1 = NamedNode("1")
    node2 = NamedNode("2")

    connect(root_up, node1)
    connect(root_up, node2)
    print(list(root_up.leaves()))
    try:
        print(root_up.top())
    except AssertionError as e:
        print("AssertionError", e)
    top = NamedNode("Top")
    for leaf in root_up.leaves():
        connect(leaf, top)
    print(root_up.top())
    print(list(root_up.leaves()))

    for node in root_up.iterate():
        print(node)

    for node in root_up.iterate_as_tree():
        print(node)


if __name__ == "__main__":
    test_dag_iteration()



import numpy as np

from matplotlib import pyplot as plt

import torch
import torchvision
import torchvision.transforms as transforms

import PIL
from PIL import Image
import scipy.misc
import cv2
import os


# all images are in RGB by default
# last dimension is color by default (1 or 3). In pytorch - this is the first dimension


def cast_numpy(im):
    if isinstance(im, np.ndarray):
        if len(im.shape) == 2:
            return im
        elif len(im.shape) == 3:
            if im.shape[-1] not in [1, 3] and im.shape[0] in [1, 3]:
                im = np.transpose(im, (1, 2, 0))
        else:
            raise Exception("Not an image")
        return im
    elif isinstance(im, PIL.Image.Image):
        return np.asarray(im)
    elif isinstance(im, torch.Tensor):
        return cast_numpy(im.numpy())
    elif isinstance(im, str) and os.path.exists(im):
        im = cv2.imread(im)
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
        return im
    else:
        raise Exception("Unknown image type")


def cast_pytorch_image(im):
    if not isinstance(im, np.ndarray):
        im = cast_numpy(im)

    if len(im.shape) != 3:
        raise NotImplementedError("Not implemented")

    if im.shape[0] not in [1, 3]:
        if im.shape[-1] in [1, 3]:
            im = np.transpose(im, (2, 0, 1))
        else:
            raise Exception("Not an image")
    return im


def cast_image(im):
    if isinstance(im, np.ndarray) and len(im.shape) == 3 and im.shape[0] == 1:
        im = im[0]
    else:
        im = cast_numpy(im)
    return Image.fromarray(im)


def cast_grayscale(im):
    if len(im.shape) > 2 and im.shape[2] == 3:
        im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
    else:
        #         im = cv2.imdecode(im, cv2.CV_LOAD_IMAGE_GRAYSCALE)
        return im
    return im


def show(im):
    im = cast_numpy(im)
    if len(im.shape) == 2:
        plt.imshow(im, cmap="gray")
    else:
        plt.imshow(im)


def save(im, fn):
    scipy.misc.imsave(fn, im)


def resize(im, new_shape, interpolation=cv2.INTER_CUBIC):
    im = cv2.resize(im, new_shape, interpolation=interpolation)
    return im


def flip(im):
    return cv2.flip(im, 1)



import sys
import math
import stackprinter
from utils_ak.coder import json_coder
from loguru import logger
import better_exceptions


def _get_stack(exception, engine="better_exceptions"):
    assert engine in ["stackprinter", "better_exceptions"]

    if engine == "better_exceptions":
        return "".join(better_exceptions.format_exception(*exception))
    elif engine == "stackprinter":
        return stackprinter.format(exception)


def format_as_json(record):
    assert "_json" not in record["extra"]

    extra = dict(record["extra"])
    extra.pop("source", None)

    record_dic = {
        "level": record["level"].name,
        "message": record["message"],
        "ts": int(math.floor(record["time"].timestamp() * 1000)),  # epoch millis
        "inner_source": record["extra"].get("source", ""),
        "extra": extra,
        "stack": "",
        "error": "",
    }

    if record["exception"]:
        record_dic["stack"] = _get_stack(record["exception"])
        record_dic["error"] = record_dic["stack"].split("\n")[-1]

    record["extra"]["_json"] = json_coder.encode(record_dic)
    return "{extra[_json]}\n"


def time_formatter(record):
    return "<green>{time:YYYY-MM-DD HH:mm:ss!UTC}</green>"


def level_formatter(record):
    return "<level>{level: <8}</level>"


def name_formatter(record):
    return "<cyan>{name}</cyan>"


def message_formatter(record):
    return "<level>{message}</level>"


def module_formatter(record):
    return "<cyan>{module}</cyan>"


def extra_formatter(record, fancy=True):
    if fancy:

        values = []
        for i, (k, v) in enumerate(record["extra"]["_extra"].items()):
            values.append(
                ": ".join(
                    [
                        f"<magenta>{{extra[_extra_{2 * i}]}}</magenta>",
                        f"<yellow>{{extra[_extra_{2 * i + 1}]}}</yellow>",
                    ]
                )
            )
            record["extra"][f"_extra_{2 * i}"] = k
            record["extra"][f"_extra_{2 * i + 1}"] = v
        return ", ".join(values)

    return "<yellow>{extra[_extra]}</yellow>"


def exception_formatter(record):
    if not record["exception"]:
        return

    assert all(key not in record["extra"] for key in ["_stack", "_error"])

    record["extra"]["_stack"] = _get_stack(record["exception"])
    record["extra"]["_error"] = record["extra"]["_stack"].split("\n")[-1]
    _extra = dict(record["extra"])
    _extra.pop("_stack")
    _extra.pop("_error")
    record["extra"]["_extra"] = _extra
    return "<red>{extra[_error]}</red>\n<red>{extra[_stack]}</red>"


def format_with_trace(
    record,
    formatters=(
        time_formatter,
        module_formatter,
        message_formatter,
        extra_formatter,
        exception_formatter,
    ),
    separator=" | ",
):
    # workaround for custom extra
    record["extra"]["_extra"] = dict(record["extra"])
    values = [formatter(record) for formatter in formatters]
    values = [value for value in values if value]
    return separator.join(values) + "\n"


def configure_loguru_stdout(
    level="DEBUG",
    remove_others=True,
    formatter=format_with_trace,
):
    if remove_others:
        logger.remove()

    logger.add(
        sys.stdout,
        level=level,
        format=formatter,
    )


def test():
    configure_loguru_stdout(formatter=format_as_json)

    logger.info("Info message", foo="bar")
    a = 1
    b = 0
    try:
        a / b
    except ZeroDivisionError:
        logger.error("Oups...")
        logger.exception("Oups...")

    configure_loguru_stdout(formatter=format_with_trace)
    logger.info("Info message", foo="bar")

    try:
        a / b
    except ZeroDivisionError:
        logger.error("Oups...")
        logger.exception("Oups...")


if __name__ == "__main__":
    test()



from utils_ak.fluid_flow import *


def test_flow_hub_1():
    parent = Container("Parent", value=100, max_pressures=[None, 20])

    hub = Hub("Hub")

    child1 = Container("Child1", max_pressures=[15, None])
    child2 = Container("Child2", max_pressures=[10, None])

    pipe_connect(parent, hub, "parent-hub")
    pipe_connect(hub, child1, "hub-child1")
    pipe_connect(hub, child2, "hub-child2")

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_hub_2():
    parent = Container("Parent", value=100, max_pressures=[None, 20])

    hub = Hub("Hub")

    child1 = Processor("Child1", max_pressures=[15, None], limits=[30, None])
    child2 = Processor("Child2", max_pressures=[10, None])

    pipe_connect(parent, hub, "parent-hub")
    pipe_connect(hub, child1, "hub-child1")
    pipe_connect(hub, child2, "hub-child2")

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_hub_3():
    parent = Container("Parent", item="a", value=100, max_pressures=[None, 20])

    hub = Hub("Hub")

    child1 = Container("Child1", item="b", max_pressures=[15, None])
    child2 = Container("Child2", item="a", max_pressures=[10, None])

    pipe_connect(parent, hub, "parent-hub")
    pipe_connect(hub, child1, "hub-child1")
    pipe_connect(hub, child2, "hub-child2")

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


if __name__ == "__main__":
    # test_flow_hub_1()
    # test_flow_hub_2()
    test_flow_hub_3()



""" Dictionary with dot as a level separator. """
from utils_ak.coder import json_coder


def is_dict_like(obj):
    """
    Return if provided object is type of dict or instance of DotDict class
    """
    return isinstance(obj, dict) or isinstance(obj, dotdict)


class dotdict(dict):
    def __init__(self, dic=None, as_default_dict=False):
        super().__init__()

        if dic is not None and not is_dict_like(dic):
            raise Exception("Bad dic input")

        self._as_default_dict = as_default_dict

        if dic is not None:
            for k, v in dic.items():
                if is_dict_like(v):
                    self[k] = self._cls(v)
                else:
                    self[k] = v

    @property
    def sys_attrs(self):
        return ["_as_default_dict", "_dic"]

    def __delattr__(self, item):
        return self.__delitem__(item)

    def _cls(self, dic=None):
        """ Inherit parent class. """
        return dotdict(dic, as_default_dict=self._as_default_dict)

    def __setattr__(self, key, value):
        if key not in self.sys_attrs:
            self[key] = value
        else:
            super().__setattr__(key, value)

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __getitem__(self, key):
        if key in self.sys_attrs:
            return super().__getitem__(key)

        if key in self or self._as_default_dict:
            return self.get(key)
        raise KeyError(key)

    def __setitem__(self, key, val):
        self._set(key, val)

    def __contains__(self, key):
        split_keys = self._split_key(key)

        first = split_keys[0]

        if len(split_keys) == 1:
            return super().__contains__(key)

        if not super().__contains__(first) or not is_dict_like(self[first]):
            return False

        return ".".join(split_keys[1:]) in self[first]

    def _set(self, key, value):
        obj = self

        split_keys = self._split_key(key)

        for key in split_keys[:-1]:
            obj = obj.setdefault(key, self._cls())
            if not is_dict_like(obj):
                raise Exception("Cannot assign new value, internal obj is not dict")

        if len(split_keys) == 1:
            super().__setitem__(split_keys[-1], value)
        else:
            obj._set(split_keys[-1], value)

        # obj.__setitem__(split_keys[-1], value)

    def delete(self, key):
        """
        Delete provided compound key from `dotdict`
        """
        obj = self
        split_keys = key.split(".")
        for key in split_keys:
            if key == split_keys[-1]:
                del obj[key]
                break
            if isinstance(obj, dotdict):
                obj = self.__getitem__(key)
            else:
                obj = obj.__getitem__(key)

    def get(self, key, default=None):
        """
        Get value for provided compound key. In a case of accessed value of a list type returns its first element.
        """
        split_keys = self._split_key(key)
        first = split_keys[0]

        if first in self:
            obj = super().__getitem__(first)

            if first == key:
                # no dotted notation used
                return obj

            if is_dict_like(obj):
                return obj.get(".".join(split_keys[1:]))
            else:
                return default
        else:
            if self._as_default_dict:
                return self.setdefault(first, self._cls())
            else:
                return default

    def __getstate__(self):
        return dict(self)

    def copy(self):
        return self._cls(self)

    @staticmethod
    def _split_key(key):
        if hasattr(key, "split"):
            return key.split(".")
        return [key]


def test():
    dd = dotdict({0: 1})

    print(dd[0])

    dd = dotdict({"a": {"b": 1, "c": [1, 2]}}, as_default_dict=False)
    print(dd)
    print(dict(dd))
    print(dd["a.c"])
    print(dd.get("a.c.does_not_exist"))

    try:
        dd["a.c.does_not_exist"]
    except KeyError as e:
        print("error", e)

    dd["x.y.z"] = 1
    print(dd)
    print(dd.x.y.z)

    import collections

    print(isinstance(dd, collections.Mapping))

    print(dd.a["c"])

    """
    Output:
    {'a': {'b': 1, 'c': [1, 2]}}
        {'a': {'b': 1, 'c': [1, 2]}}
        [1, 2]
        None
        error 'c'
        {'a': {'b': 1, 'c': [1, 2]}, 'x': {'y': {'z': 1}}}
        1
        True
        [1, 2]
        """

    import pickle

    print(pickle.dumps(dd))

    dd = dotdict(as_default_dict=True)
    dd["d"]["e"]["f.g"] = 1
    print(dd)

    dd.foo.bar = 2
    print(dd)

    print(json_coder.encode(dd))

    from collections import Mapping

    print(isinstance(dd, Mapping))

    import pickle

    print(pickle.dumps(dd))


if __name__ == "__main__":
    test()



# online/offline
# cache for state
# state - подается на вход?

# один processor на один топик? Да, но client снаружи - слушает сообщения
# Как безопасно хранить state? В чем проблема хранить в файле? Слишком часто обновлять и перезаписывать файл.

from .kafka_client import KafkaClient


class KafkaProcessor:
    def __init__(self, topic, bootstrap_servers="localhost:9092"):
        pass



import logging

# todo: migrate to loguru


class ZMQLogHandler(logging.Handler):
    def __init__(self, publisher, root_topic=None):
        logging.Handler.__init__(self)
        self.publisher = publisher
        self.root_topic = root_topic

    def emit(self, record):
        topic_list = [record.levelname]
        if self.root_topic:
            topic_list = [self.root_topic] + topic_list
        result_topic = ".".join(topic_list)
        result_msg = self.format(record)
        self.publisher.publish(result_topic, result_msg)



from utils_ak.clock import *


def cast_simple_vector(obj):
    if isinstance(obj, SimpleVector):
        return obj
    elif isinstance(obj, list):
        return SimpleVector(obj)
    elif isinstance(obj, tuple):
        # todo: del
        # return SimpleVector(list(obj))
        return SimpleVector(obj)
    elif isinstance(obj, int):
        return SimpleVector([0] * obj)
    else:
        raise Exception("Unknown object type")


def cast_vector_values(obj):
    if isinstance(obj, SimpleVector):
        return obj.values
    elif isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return obj
    elif isinstance(obj, int):
        return [0] * obj
    else:
        raise Exception("Unkown object type")


class SimpleVector:
    def __init__(self, values):
        self.values = cast_vector_values(values)
        self.n_dims = len(self.values)

    def __add__(self, other):
        return SimpleVector([self[i] + other[i] for i in range(self.n_dims)])

    def __sub__(self, other):
        return SimpleVector([self[i] - other[i] for i in range(self.n_dims)])

    def __getitem__(self, item):
        return self.values[item]

    def __repr__(self):
        return "SimpleVector({})".format(self.values)

    def __str__(self):
        return "SimpleVector({})".format(self.values)

    def __setitem__(self, index, value):
        self.values[index] = value

    def add(self, other):
        for i, v in enumerate(other.values):
            self.values[i] += v
        return self

    def __iter__(self):
        return iter(self.values)


def test_simple_vector():
    v1 = SimpleVector([1, 2])
    v2 = SimpleVector([3, 5])
    print(v1 + v2, v1 - v2)


if __name__ == "__main__":
    test_simple_vector()



import subprocess
from loguru import logger

from utils_ak.str import cast_unicode


def execute(cmd, is_async=False, verbose=True):
    if verbose:
        logger.trace("Executing command", cmd=cmd, is_async=is_async)

    if is_async:
        return subprocess.Popen(cmd, shell=True)
    else:
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        output, error = process.communicate()
        res = " ".join([cast_unicode(x) for x in [output, error] if x])
        if verbose:
            logger.trace("Execution response", response=res)
        return res



import time
import asyncio
import itertools

from utils_ak.callback_timer import *
from utils_ak.architecture.func import *
from utils_ak.coder import *
from utils_ak.message_queue import *
from utils_ak.str import *

from loguru import logger as global_logger


TIME_EPS = 0.001


class SimpleMicroservice:
    """ Microservice base class with timers and subscriber. Works on asyncio. """

    def __init__(
        self,
        id,
        message_broker,
        logger=None,
        coder=None,
    ):
        self.id = id

        # aio
        self.tasks = []

        # sync
        self.timers = []

        # {collection: callback}
        self.callbacks = {}

        self.message_broker = message_broker
        self.broker = cast_message_broker(message_broker)

        # {broker: collection::topic}
        self.subscribed_to = {}

        self.logger = logger or global_logger
        self.logger = self.logger.bind(microservice_id=str(id))

        self.default_exception_timeout = 10.0
        self.max_exception_timeout = 3600
        self.fail_count = 0

        self.is_active = True

        # todo: make JsonCoder default after datetime can be decoded properly
        self.coder = coder or MsgPackCoder()

    def stop(self):
        self.logger.info("Stopping microservice")
        self.is_active = False

    def register_publishers(self, collections):
        if isinstance(self.broker, ZMQBroker):
            for collection in collections:
                self.add_timer(
                    self.publish,
                    interval=1,
                    n_times=1,
                    args=(collection, "register"),
                )

    def _args_formatter(self, topic, msg):
        # convert msg to kwargs
        return (cast_unicode(topic),), self.coder.decode(msg)

    def add_timer(self, *args, **kwargs):
        """
        :param interval: timer will be run every interval
        :param callback: func
        """
        self.timers.append(CallbackTimer(*args, **kwargs))

    def add_schedule(self, *args, **kwargs):
        """
        :param pattern: cron-like pattern with seconds:
              sec min hour monthday month weekday
              *    5    *     *       *     *     -> run every 5th minute of every hour
        :param callback:
        :param init_run: bool
        """
        self.timers.append(ScheduleTimer(*args, **kwargs))

    def subscribe(self, collection, topic):
        self.broker.subscribe(collection, topic)

    def add_callback(
        self,
        collection,
        topic,
        callback=None,
        formatter="default",
        filter=None,
        topic_formatter=cast_unicode,
        subscribe=True,
    ):
        if subscribe:
            self.broker.subscribe(collection, topic)
        self._add_callback(
            collection, topic, callback, formatter, filter, topic_formatter
        )

    def _add_callback(
        self,
        collection,
        topic,
        callback=None,
        formatter="default",
        filter=None,
        topic_formatter=cast_unicode,
    ):
        assert isinstance(topic, str), "Topic must be a str"

        if formatter == "default":
            formatter = self._args_formatter

        handler = self.callbacks.setdefault(collection, PrefixHandler())
        handler.set_topic_formatter(topic_formatter)
        handler.add(topic, callback=callback, filter=filter, formatter=formatter)
        return handler

    def publish_raw(self, collection, topic, msg):
        self.broker.publish(collection, topic, msg)

    def publish(self, collection, topic, **msg):
        self.publish_raw(collection, topic, self.coder.encode(msg))

    def wrap_coroutine_timer(self, timer):
        async def f():
            while True:
                try:
                    ran = await timer.run_if_possible_async()

                    if ran and self.fail_count:
                        self.logger.debug("Success. Resetting the failure counter")
                        self.fail_count = 0

                except Exception as e:
                    self.on_exception(e, "Exception occurred at the timer callback")

                if not self.is_active:
                    return

                await asyncio.sleep(max(timer.next_call - time.time() + TIME_EPS, 0))

        return f()

    def wrap_broker_coroutine(self, timeout=0.01):
        async_supported = self.broker.async_supported

        async def f():
            while True:
                try:
                    if async_supported:
                        received = await self.broker.aiopoll(timeout=1.0)
                    else:
                        received = self.broker.poll(timeout)

                    if received:
                        for record in received:
                            collection, topic, msg = record

                            try:
                                self.logger.trace(
                                    f"Received new message",
                                    topic=str(topic),
                                    msg=str(msg),
                                )
                                await self.callbacks[collection].call_async(topic, msg)

                                if self.fail_count != 0:
                                    self.logger.info(
                                        "Success. Resetting the failure counter"
                                    )
                                    self.fail_count = 0

                            except Exception as e:
                                self.on_exception(
                                    e, "Exception occurred at the callback"
                                )
                except Exception as e:
                    self.on_exception(e, "Failed to receive the message")

                if not self.is_active:
                    return

                await asyncio.sleep(0)

        return f()

    def _run_async(self):
        self.loop = asyncio.get_event_loop()
        self.logger.info("Microservice started")

        for timer in self.timers:
            self.tasks.append(self.wrap_coroutine_timer(timer))

        self.tasks.append(self.wrap_broker_coroutine())

        self.tasks = [asyncio.ensure_future(task) for task in self.tasks]
        self.loop.run_until_complete(asyncio.wait(self.tasks))

    def run(self):
        return self._run_async()

    def on_exception(self, e, msg):
        self.logger.exception("Generic microservice error")
        to_sleep = min(
            self.default_exception_timeout * 2 ** (self.fail_count - 1),
            self.max_exception_timeout,
        )
        time.sleep(to_sleep)
        self.fail_count += 1



import decimal
import datetime as datetime_module


def _test_coder(coder, assert_equal=True):
    obj = {
        "int": 1,
        "datetime": datetime_module.datetime.now(),
        "decimal": decimal.Decimal("0.12431234123000"),
        "date": datetime_module.date(2018, 1, 1),
        "time": datetime_module.datetime.now().time(),
    }

    encoded = coder.encode(obj)
    decoded = coder.decode(encoded)
    print(encoded)
    print(decoded)
    if assert_equal:
        assert obj == decoded



from utils_ak.fluid_flow.actor import Actor


class Stub(Actor):
    def __init__(self, name=None):
        super().__init__(name)

    def update_pressure(self, ts):
        for pipe in self.parents:
            pipe.pressures["in"] = 0
        for pipe in self.children:
            pipe.pressures["out"] = 0

    def __str__(self):
        return f"Stub {self.name}"



import os
import sys
import random
import re
import math
import numpy as np
import pandas as pd
import anyconfig
import copy
import pyperclip
import glob
import logging
import uuid
import openpyxl


import inspect
from datetime import datetime, timedelta
from pprint import pprint
from bson.objectid import *
from loguru import logger
from icecream import ic


from .builtin import *
from .callback_timer import *
from .clock import *
from .dict import *
from .exceptions import *

# from .git import *
from .numeric import *
from .os import *
from .pandas import *
from .re import *
from .reflexion import *
from .str import *
from .time import *
from .tqdm import *
from .jupyter import *
from .simple_microservice import *

# from .config import *
from .streaming import *
from .properties import *
from .naming import *
from .simple_event_manager import *
from .split_file import *
from .block_tree import *
from .portion import *
from .color import *
from .openpyxl import *
from .dag import *
from .iteration import *
from .fluid_flow import *
from .message_queue import *
from .deployment import *
from .loguru import *



from utils_ak.clock import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import PipeMixin
from utils_ak.fluid_flow.calculations import *
import pandas as pd


class Container(Actor, PipeMixin):
    def __init__(self, name, value=0, item="default", max_pressures=None, limits=None):
        super().__init__(name)
        self.item = item
        self.value = value

        self.df = pd.DataFrame(
            index=["in", "out"], columns=["max_pressure", "limit", "collected"]
        )
        self.df["max_pressure"] = max_pressures
        self.df["limit"] = limits
        self.df["collected"] = 0.0

        self.transactions = []

    def is_limit_reached(self, orient):
        if (
            self.df.at[orient, "limit"]
            and abs(self.df.at[orient, "collected"] - self.df.at[orient, "limit"])
            < ERROR
        ):
            return True
        return False

    def update_value(self, ts, factor=1):
        if self.last_ts is None:
            return

        self.add_value(ts, "in", (ts - self.last_ts) * self.speed("in") * factor)
        self.add_value(ts, "out", -(ts - self.last_ts) * self.speed("out"))

    def add_value(self, ts, orient, value):
        if not value:
            return
        self.value += value
        self.df.at[orient, "collected"] += abs(value)
        self.transactions.append([self.last_ts, ts, value])

    def active_periods(self, orient="in"):
        if not self.transactions:
            return []
        return [[self.item, self.transactions[0][0], self.transactions[-1][1]]]

    def update_pressure(self, ts, orients=("in", "out")):
        for orient in orients:
            if self.pipe(orient):
                pressure = (
                    self.df.at[orient, "max_pressure"]
                    if not self.is_limit_reached(orient)
                    else 0
                )
                self.pipe(orient).set_pressure(orient, pressure, self.item)

    def update_speed(self, ts, set_out_pressure=True):
        input_speed = self.speed("in")

        if set_out_pressure:
            if self.pipe("out") and abs(self.value) < ERROR:
                self.pipe("out").set_pressure(
                    "out",
                    nanmin([self.pipe("out").pressures["out"], input_speed]),
                    self.item,
                )

    def reset(self):
        super().reset()
        self.transactions = []

    def update_triggers(self, ts):
        values = []
        if self.drain() < 0:
            values.append(["empty_container", self.value, self.drain()])

        for orient in ["in", "out"]:
            if self.df.at[orient, "limit"]:
                values.append(
                    [
                        f"{orient} limit",
                        self.df.at[orient, "limit"] - self.df.at[orient, "collected"],
                        self.speed(orient),
                    ]
                )

        values = [
            value for value in values if value[1] > ERROR and abs(value[2]) > ERROR
        ]
        etas = [value[1] / abs(value[2]) for value in values]
        if etas:
            self.add_event("update.trigger", ts + min(etas), {})

    def __str__(self):
        return f"Container {self.name}:{self.item}"

    def stats(self):
        return {"value": self.value}

    def display_stats(self):
        return self.value



import fnmatch
import shutil
from contextlib import contextmanager
from _sha512 import sha512
from uuid import uuid4
import os
import tempfile as tmp
import subprocess
import os
import platform


from utils_ak.str import cast_bytes


def make_directories(path):
    dirname = os.path.dirname(path)
    if not dirname:
        return
    if not os.path.exists(dirname):
        os.makedirs(dirname)


makedirs = make_directories


def remove_path(path):
    if not os.path.exists(path):
        return
    elif os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path, ignore_errors=True)
    else:
        raise Exception("Path {} is not a file or a dir".format(path))


def rename_path(path1, path2):
    os.rename(path1, path2)


def copy_path(path_from, path_to):
    shutil.copy(path_from, path_to)


def touch_file(fn):
    with open(fn, "w") as f:
        pass


def remove_empty_directories(path, remove_root=True):
    """Function to remove empty folders"""
    if not os.path.isdir(path):
        return

    # remove empty subfolders
    fns = os.listdir(path)

    for fn in fns:
        fullpath = os.path.join(path, fn)
        if os.path.isdir(fullpath):
            remove_empty_directories(fullpath)

    # if folder empty, delete it
    fns = os.listdir(path)

    if not fns and remove_root:
        os.rmdir(path)


def list_files(path, pattern=None, recursive=True):
    if not recursive:
        fns = os.listdir(path)
    else:
        # glob.glob('**/*') is slower 2.5 times than simple os.walk. It also returns directories
        fns = []
        for root, dirs, files in os.walk(path):
            # todo: make os.path.join faster?
            fns += [os.path.join(root, fn) for fn in files]
    if pattern:
        fns = [fn for fn in fns if fnmatch.fnmatch(fn, pattern)]
    return fns


def gen_fn_hash(s):
    return sha512(s).hexdigest()


def gen_tmp_fn(s=None, extension=None):
    if not s:
        s = uuid4().hex
    s = cast_bytes(s, encoding="utf-8")
    res = os.path.join("/tmp/", gen_fn_hash(s))
    if extension:
        res = res + extension
    return res


@contextmanager
def tempfile(*args, **kwargs):
    """Context for temporary file.

    Will find a free temporary filename upon entering
    and will try to delete the file on leaving, even in case of an exception.

    Parameters
    ----------
    suffix : string
        optional file suffix
    dir : string
        optional directory to save temporary file in
    """

    tf = tmp.NamedTemporaryFile(delete=False, *args, **kwargs)
    tf.file.close()
    try:
        yield tf.name
    finally:
        try:
            os.remove(tf.name)
        except OSError as e:
            if e.errno == 2:
                pass
            else:
                raise


@contextmanager
def open_atomic(filepath, *args, **kwargs):
    """Open temporary file object that atomically moves to destination upon
    exiting.

    Allows reading and writing to and from the same filename.

    The file will not be moved to destination in case of an exception.

    Parameters
    ----------
    filepath : string
        the file path to be opened
    fsync : bool
        whether to force write the file to disk
    *args : mixed
        Any valid arguments for :code:`open`
    **kwargs : mixed
        Any valid keyword arguments for :code:`open`
    """
    fsync = kwargs.pop("fsync", False)

    with tempfile(dir=os.path.dirname(os.path.abspath(filepath))) as tmppath:
        with open(tmppath, *args, **kwargs) as file:
            try:
                yield file
            finally:
                if fsync:
                    file.flush()
                    os.fsync(file.fileno())
        os.rename(tmppath, filepath)


def getenv_boolean(var_name, default_value=False):
    result = default_value
    env_value = os.getenv(var_name)
    if env_value is not None:
        result = env_value.upper() in ("TRUE", "1")
    return result


def open_file_in_os(fn):
    fn = os.path.abspath(fn)
    if platform.system() == "Darwin":  # macOS
        subprocess.call(("open", fn))
    elif platform.system() == "Windows":  # Windows
        os.startfile(fn)
    else:  # linux variants
        subprocess.call(("xdg-open", fn))


if __name__ == "__main__":
    print(gen_tmp_fn(extension=".operation.canary"))



import time
import pandas as pd
import json


# todo: create subclock - for prefixes
# todo: remove pandas dependency
# todo: auto trie generation? Generate subtree while possible
# todo: remove overhead from timing. Is it even possible?


class Clock(object):
    """ Time measurement of any code piece. """

    def __init__(
        self,
        checkpoints=None,
        laps=None,
        auto_close=False,
        sep="/",
        max_observations=100000,
        enabled=True,
    ):
        self.checkpoints = checkpoints or {}
        self.laps = laps or {}

        self.auto_close = auto_close
        self.last_key = None

        self.prefix = None
        self.sep = sep

        self.max_observations = max_observations

        self.enabled = enabled

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def set_prefix(self, prefix=None):
        if not prefix:
            self.prefix = None
        else:
            prefix = prefix if isinstance(prefix, list) else [prefix]
            self.prefix = self.sep.join(prefix)

    def add_lap(self, key, lap):
        self.laps.setdefault(key, []).append(lap)
        if len(self.laps[key]) > self.max_observations:
            self.laps[key] = self.laps[key][-self.max_observations :]

    def clock(self, key="clock", reset=False):
        if not self.enabled:
            return

        if isinstance(key, list):
            key = self.sep.join(key)

        if self.prefix:
            key = self.sep.join([self.prefix, key])

        if reset:
            self.pop(key)
        t = time.perf_counter()
        key = str(key)

        if key not in self.checkpoints:
            self.checkpoints[key] = t

            if self.auto_close and self.last_key:
                self.add_lap(self.last_key, t - self.checkpoints.pop(self.last_key))
        else:
            self.add_lap(key, t - self.checkpoints.pop(key))
        self.last_key = key
        return ContextClockWithKey(key, self)

    def stop(self, key=None, raise_if_not_started=True):
        """
        :param key: str.
        :param raise_if_not_started:
        :return: bool. Raise if exists
        """
        if not self.enabled:
            return

        t = time.perf_counter()
        if not key:
            for key, t0 in self.checkpoints.items():
                self.add_lap(key, t - t0)
            self.checkpoints = {}
        else:
            if key in self.checkpoints:
                self.add_lap(key, t - self.checkpoints.pop(key))
            elif raise_if_not_started:
                raise Exception("Key {} is not started".format(key))

    def start(self, key, raise_if_already_started=True):
        """
        :param key: str.
        :param raise_if_already_started:  bool. Raise if exists
        :return:
        """
        if not self.enabled:
            return

        if key not in self.checkpoints:
            return self.clock(key)
        elif raise_if_already_started:
            raise Exception("Key {} has already started".format(key))

    def stats(self):
        keys = list(self.laps.keys())
        if not keys:
            return pd.DataFrame(index=["count", "mean", "sum", "max", "min"])
        df = pd.concat(
            [pd.DataFrame({k: self.laps[k]}) for k in keys], ignore_index=True, axis=1
        )
        df.columns = keys
        stats_df = df.agg(["count", "mean", "sum", "max", "min"])
        return stats_df

    def pop(self, key):
        self.checkpoints.pop(key, None)
        self.laps.pop(key, None)

    def reset(self):
        self.checkpoints = {}
        self.laps = {}

    def print_trie(self, by="sum", sep=None, precision=3):
        sep = sep or self.sep

        import pygtrie

        t = pygtrie.StringTrie(separator=sep)

        stats = self.stats().T[[by]]
        for key, row in stats.sort_index().iterrows():
            t[key] = row[by]

        def is_child(key):
            return not t.has_subtrie(key)

        def depth_print(s, depth=0):
            print("\t" * depth + s)

        def print_trie(path_conv, path, children, value=None):
            path = path_conv(path)
            depth = path.count(sep)

            if t.has_key(path):
                value_repr = value if not precision else round(value, precision)
                path_repr = path.split("/")[-1]
                if is_child(path):
                    # child
                    depth_print("{}: {}".format(path_repr, value_repr), depth)
                else:
                    depth_print("{}: {}".format(path_repr, value_repr), depth)
            for child in children:
                pass
            return value

        def add_other(path_conv, path, children, value=None):
            path = path_conv(path)
            if value and not is_child(path):
                other_key = "/".join([path, "other"])
                if not t.has_key(other_key):
                    t[other_key] = value - sum(child for child in children)
            else:
                for child in children:
                    pass
            return value

        if by == "sum":
            t.traverse(add_other)
        t.traverse(print_trie)

    def __call__(self, *args, **kwargs):
        return self.clock(*args, **kwargs)

    def __getitem__(self, item):
        return self.stats().T["sum"][item]

    @property
    def state(self):
        return {"laps": self.laps}

    def save(self, fn):
        with open(fn, "w") as f:
            json.dump(self.state, f, indent=1)

    def load(self, fn):
        with open(fn, "r") as f:
            js = json.load(f)
            self.laps = js["laps"]


clock = Clock(enabled=False)


class ContextClockWithKey:
    def __init__(self, key, clock=clock):
        self.clock = clock
        self.key = key

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clock.stop(self.key)


from functools import wraps


def clockify(key=None):
    def _clockify(f):
        @wraps(f)
        def inner(*args, **kwargs):

            _values = [str(f.__module__), f.__name__]
            if key:
                _values.append(key)
            _key = "/".join(_values)

            clock.start(_key)
            try:
                res = f(*args, **kwargs)
                clock.stop(_key)
            except:
                clock.stop(_key)
                raise
            return res

        return inner

    return _clockify


@clockify("time")
def test_time():
    time.sleep(1)


@clockify("time_with_exception")
def test_time_with_exception():
    time.sleep(1)
    raise Exception("test")


# todo: make a reproducible test
def test():
    # usage 1
    clock = Clock()

    test_time()
    try:
        test_time_with_exception()
    except:
        pass
    print(clock.stats())
    clock.reset()

    # start clock
    clock.clock("first")
    time.sleep(0.1)
    # stop clock
    clock.clock("first")

    clock.clock("second")
    time.sleep(0.2)
    clock.clock("second")

    # this will add statistics to the 'first' key
    clock.clock("first")
    time.sleep(0.1)
    clock.clock("first")

    print(clock.stats())

    # usage 2
    clock = Clock(auto_close=True)
    clock.clock("first")
    time.sleep(0.1)
    # this will auto-close 'first' clock
    clock.clock("second")
    time.sleep(0.2)

    # this stops all clocks
    clock.stop()
    print("Auto close")
    print(clock.stats())

    clock = Clock()
    # trie example
    clock.reset()
    clock.clock("a")

    for i in range(10):
        clock.clock("a/b")
        time.sleep(0.1)
        clock.clock("a/b")
    clock.clock("a")
    print(clock.stats())

    try:
        clock.print_trie()
        clock.print_trie(by="count")
    except:
        pass

    print(clock["a"])

    clock.save("clock.json")

    clock.reset()
    print(clock.stats())
    clock.load("clock.json")
    print(clock.stats())

    # check max observations
    clock = Clock(max_observations=3)
    for i in range(10):
        print(clock.laps.get("a"))
        clock.start("a")
        time.sleep(0.01 * i)
        clock.stop("a")

    with clock("asdf"):
        time.sleep(1)

    print(clock.stats())


if __name__ == "__main__":
    test()



""" Git functionality. """
import glob
import os
import sys
from collections import defaultdict

import git
import traceback

from utils_ak.time.dt import cast_datetime
from utils_ak.re import search_one


def is_git_repo(path):
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False


def add_git_repos_to_path(path, recursive=False):
    if not recursive:
        paths = glob.glob(os.path.join(path, "*/"))
    else:
        paths = glob.glob(os.path.join(path, "**/*/"), recursive=True)

    paths = [path for path in paths if is_git_repo(path)]

    for path in paths:
        sys.path.append(path)


def git_pull(path, full_output=False):
    if not is_git_repo(path):
        return {"status": "fail", "e": "Not a git repo"}

    res = defaultdict(str)
    try:
        g = git.cmd.Git(path)
        # if stash:
        #     stash_output = g.stash()
        #     res['output'] += stash_output + '\n'

        output = g.pull()
        res["output"] += output + "\n"

        # if stash:
        #     try:
        #         unstash_output = g.stash('apply')
        #         res['output'] += unstash_output + '\n'
        #     except:
        #         pass

        if "Already up-to-date" in output:
            res["status"] = "Already up-to-date"
        else:
            res["status"] = "success"

    except Exception as e:
        res["status"] = "fail"
        res["e"] = traceback.format_exc()
    res = dict(res)
    if not full_output:
        res = res["status"]
    return res


def git_pull_many(path, repos=None, recursive=False):
    if not recursive:
        paths = glob.glob(os.path.join(path, "*/"))
    else:
        paths = glob.glob(os.path.join(path, "**/*/"), recursive=True)
    paths = [path for path in paths if is_git_repo(path)]
    if repos:
        paths = [
            path for path in paths if os.path.basename(os.path.dirname(path)) in repos
        ]
    return {path: git_pull(path) for path in paths}


def get_commit_info(path):
    repo = git.Repo(path)
    commit_info = repo.git.show()

    commit_dic = {
        "hash": search_one(r"commit (.+)\n", commit_info),
        "date": search_one(r"Date:\s+(.+)\n", commit_info),
        "author": search_one(r"Author:\s+(.+)\n", commit_info),
        "full_info": commit_info,
    }
    try:
        commit_dic["date"] = cast_datetime(commit_dic["date"])
    except:
        pass
    return commit_dic


if __name__ == "__main__":
    # res = git_pull_many(r'C:\Users\xiaomi\YandexDisk\IT\Python\pycharm')
    # res = git_pull_many('/home/jentos/Projects/quantribution')
    # pprint(res)
    print(get_commit_info(r"C:\Users\xiaomi\YandexDisk\IT\Python\pycharm\prod"))



from utils_ak.architecture import PrefixHandler
from sortedcollections import SortedList
import logging


class SimpleEventManager:
    def __init__(self):
        self.events = SortedList(key=lambda v: v[1])  # sorted(topic, ts, event)])
        self.prefix_handler = PrefixHandler()
        self.last_ts = None

    def subscribe(self, topic, callback):
        self.prefix_handler.add(topic, callback)

    def add_event(self, topic, ts, event, duplicates_allowed=False):
        if not duplicates_allowed:
            if self.is_event_present(topic, ts, event):
                return False
        self.events.add((topic, ts, event))
        return topic, ts, event

    def is_event_present(self, topic, ts, event, ts_error=1e-5):
        if not self.events:
            return False

        ind = self.events.bisect_left([topic, ts, event])

        for increment in [1, -1]:
            cur_ind = ind
            while True:
                if cur_ind < 0 or cur_ind >= len(self.events):
                    break
                cur_event = self.events[cur_ind]

                if increment == 1 and cur_event[1] > ts + ts_error:
                    break
                if increment == -1 and cur_event[1] < ts - ts_error:
                    break

                if (
                    cur_event[0] == topic
                    and abs(cur_event[1] - ts) < ts_error
                    and cur_event[2] == event
                ):
                    return True

                cur_ind += increment

        return False

    def run(self):
        while True:
            if not self.events:
                return
            topic, ts, event = self.events.pop(0)

            if self.last_ts is not None and ts < self.last_ts:
                logging.warning("Old event was added to the events timeline")
            self.prefix_handler(topic, ts, event)
            self.last_ts = max(ts, self.last_ts or 0)


def test_simple_event_manager():
    em = SimpleEventManager()

    class Counter:
        def __init__(self):
            self.counter = 0

        def on_count(self, topic, ts, event):
            self.counter += event["num"]
            print("Current counter", topic, ts, event, self.counter)

    counter = Counter()

    em.subscribe("count", counter.on_count)

    from utils_ak.time import cast_ts
    from datetime import datetime

    now_ts = cast_ts(datetime.now())
    em.add_event("count.up", now_ts, {"num": 3})
    em.add_event("count.down", now_ts + 2, {"num": -11})

    # test is_event_prent
    assert em.is_event_present("count.up", now_ts, {"num": 3})
    assert em.is_event_present("count.up", now_ts + 1e-10, {"num": 3})
    assert em.is_event_present("count.up", now_ts - 1e-10, {"num": 3})
    assert not em.is_event_present("count.up", now_ts + 1, {"num": 3})
    assert not em.is_event_present("different topic", now_ts + 1e-10, {"num": 3})
    assert not em.is_event_present("count.up", now_ts + 1e-10, {"different event": 10})

    # test duplicate addition
    events_count_before = len(em.events)
    em.add_event("count.up", now_ts, {"num": 3})
    assert len(em.events) == events_count_before

    em.run()


if __name__ == "__main__":
    test_simple_event_manager()



from utils_ak.fluid_flow import *


def test_flow_sequence_1():
    c1 = Container("Input", value=100, max_pressures=[None, 10])

    sequence = Sequence(
        "Sequence",
        containers=[Container(str(i), max_pressures=[5, None]) for i in range(10)],
    )

    c2 = Container("Ouput", max_pressures=[None, None])

    pipe_connect(c1, sequence)
    pipe_connect(sequence, c2)
    flow = FluidFlow(c1, verbose=True)
    run_flow(flow)


def test_flow_sequence_2():
    c1 = Container("Input", value=100, max_pressures=[None, 10])

    sequence = Sequence(
        "Sequence",
        containers=[Container(str(i), max_pressures=[2 - i, None]) for i in range(2)],
    )

    c2 = Container("Ouput", max_pressures=[None, None])

    pipe_connect(c1, sequence)
    pipe_connect(sequence, c2)
    flow = FluidFlow(c1, verbose=True)
    run_flow(flow)


if __name__ == "__main__":
    # test_flow_sequence_1()
    test_flow_sequence_2()



import zmq

from utils_ak.zmq.wrappers import Subscriber, Publisher, endpoint


class ZMQClient:
    def __init__(self, context=None):
        super().__init__()
        self.poller = None
        self.context = context or zmq.Context()

        self.subscribers = {}
        self.publishers = {}

    def subscribe(self, endpoint, topic=None, conn_type="auto"):
        if not self.poller:
            self.poller = zmq.Poller()
        topic = topic or ""
        if endpoint not in self.subscribers:
            sub = Subscriber(endpoint, context=self.context, conn_type=conn_type)
            self.subscribers[endpoint] = sub
            self._register(sub)
        self.subscribers[endpoint].subscribe(topic)

    def _register(self, receiver):
        self.poller.register(receiver.socket, zmq.POLLIN)

    def publish(self, endpoint, topic, msg, conn_type="auto"):
        if endpoint not in self.publishers:
            self.publishers[endpoint] = Publisher(
                endpoint, context=self.context, conn_type=conn_type
            )
        self.publishers[endpoint].publish(topic, msg)

    def poll(self, timeout=0.0):
        if not self.subscribers:
            return

        if len(self.subscribers) == 1 and timeout == 0:
            endpoint, sub = list(self.subscribers.items())[0]
            topic, msg = sub.receive()
            return endpoint, topic, msg
        else:
            socks = dict(self.poller.poll(timeout * 1000))
            for endpoint, receiver in self.subscribers.items():
                if receiver.socket in socks and socks[receiver.socket] == zmq.POLLIN:
                    topic, msg = receiver.receive()
                    return endpoint, topic, msg



import base64
import hashlib


def safe_filename_hash(s, n=8):
    s = s.encode("utf-8")
    return (
        base64.urlsafe_b64encode(hashlib.md5(s).digest())[:n]
        .decode("utf-8")
        .replace("_", "*")
    )


def test_safe_filename_hash():
    print(safe_filename_hash("foo-bar"))


if __name__ == "__main__":
    test_safe_filename_hash()



from loguru import logger

from utils_ak.coder import cast_dict_or_list
from utils_ak.simple_vector import *
from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock


def stack_push(parent, block):
    axis = parent.props["axis"]
    cur_end = (
        0
        if not parent.children
        else max(c.y[axis] - parent.x[axis] for c in parent.children)
    )

    x = block.props.get("x", cast_simple_vector(block.n_dims))

    assert not x[axis], "Cannot push non-zero axis"

    x[axis] = cur_end
    block.props.update(x=x)
    return add_push(parent, block)


def test_stack_push():
    logger.debug("Stack push test")
    root = ParallelepipedBlock("root", n_dims=1, x=[2], axis=0)
    a = ParallelepipedBlock("a", n_dims=1, size=[4], axis=0)
    b = ParallelepipedBlock("b", n_dims=1, size=[3], axis=0)
    stack_push(root, a)
    stack_push(root, b)
    logger.debug(root)


def simple_push(parent, block, validator=None, new_props=None):
    # update props for current try
    new_props = new_props or {}
    block.props.update(**new_props)

    parent.add_child(block)

    if validator:
        try:
            validator(parent, block)
        except AssertionError as e:
            try:
                # reset parent
                parent.remove_child(block)
                # extract assertion message json
                res = cast_dict_or_list(e.__str__())  # {'disposition': 2}
                res = res or {}
                return res
            except:
                return {}
    return block


def add_push(parent, block, new_props=None):
    return simple_push(parent, block, validator=None, new_props=new_props)


def push(parent, block, push_func=stack_push, **kwargs):
    return push_func(parent, block, **kwargs)


if __name__ == "__main__":
    test_stack_push()



from matplotlib import colors


def cast_color(obj):
    if isinstance(obj, (tuple, list)):
        if all(isinstance(x, int) for x in obj):
            return "#{:02x}{:02x}{:02x}".format(*obj)
        elif all(isinstance(x, float) for x in obj):
            return colors.to_hex(obj)
    elif isinstance(obj, str):
        try:
            return cast_color(colors.to_rgb(obj))
        except:
            pass
    raise Exception("Unknown object type")


if __name__ == "__main__":
    print(cast_color("grey"))
    print(cast_color((1.0, 0.0, 1.0)))
    print(cast_color((15, 15, 16)))
    print(cast_color("#808081"))



from qset_feature_store.imports.runtime import *

from utils_ak.job_orchestrator.models import *
from utils_ak.job_orchestrator.job_orchestrator import JobOrchestrator
from utils_ak.job_orchestrator.tests.test_monitor import run_monitor
from utils_ak.job_orchestrator.tests.config.config import config


def run_job_orchestrator():
    utils.configure_loguru_stdout("DEBUG")

    connect(host=config.MONGODB_HOST, db=config.MONGODB_DB)
    logger.info("Connected to mongodb")
    controller = utils.ProcessController()

    utils.run_listener_async("job_orchestrator", message_broker=config.TRANSPORT)
    job_orchestrator = JobOrchestrator(controller, config.TRANSPORT)
    multiprocessing.Process(target=run_monitor).start()
    logger.info("Running job orchestrator...")
    job_orchestrator.run()


if __name__ == "__main__":
    run_job_orchestrator()



import multiprocessing
import time
from utils_ak.simple_microservice.microservice import SimpleMicroservice
from utils_ak.loguru import configure_loguru_stdout


class Listener(SimpleMicroservice):
    """ Logging service for current collection. """

    def __init__(self, collection, topic="", *args, **kwargs):
        super().__init__(f"Listener_{collection}", *args, **kwargs)
        self.add_callback(collection, topic, self._log)

    def _log(self, topic, **kwargs):
        self.logger.info("Listened", topic=topic, **kwargs)


def run_listener(collection, topic="", *args, **kwargs):
    """ Run listener process for current collection. """
    configure_loguru_stdout("DEBUG")
    Listener(collection, topic, *args, **kwargs).run()


def run_listener_async(collection, topic="", timeout=None, *args, **kwargs):
    """ Run listener in a separate process."""
    args = [collection, topic] + list(args)
    if timeout:
        time.sleep(timeout)
    multiprocessing.Process(target=run_listener, args=args, kwargs=kwargs).start()



from utils_ak.message_queue.brokers.broker import Broker

from utils_ak.kafka import *
from loguru import logger


class KafkaBroker(Broker):
    def __init__(self, *args, **kwargs):
        self.cli = KafkaClient(*args, **kwargs)
        self.async_supported = False  # todo: switch to True?

    def _get_kafka_topic(self, collection, topic):
        if topic:
            return f"{collection}__{topic}"
        else:
            return collection

    def _split_kafka_topic(self, kafka_topic):
        if "__" in kafka_topic:
            return kafka_topic.split("__", 1)
        else:
            return kafka_topic, ""

    def publish(self, collection, topic, msg):
        self.cli.publish(self._get_kafka_topic(collection, topic), msg)

    def subscribe(self, collection, topic, start_offset=None):
        self.cli.subscribe(
            self._get_kafka_topic(collection, topic), start_offset
        )

    def poll(self, timeout=0.0):
        response = self.cli.poll(timeout)
        if not response:
            return
        records = sum(response.values(), [])
        # todo: optimize
        return [
            list(self._split_kafka_topic(record.topic)) + [record.value]
            for record in records
        ]



from utils_ak.id_generator.id_generators import *


def test():
    for generator in [UUIDGenerator(), IncrementalIDGenerator()]:
        for i in range(5):
            print(i, generator.gen_id())


if __name__ == "__main__":
    test()



import os
import pickledb

from utils_ak.state.provider.provider import StateProvider
from utils_ak.os import *


class PickleDBStateProvider(StateProvider):
    """ I keep a number (which is a called a state). I allow for the state to not change in case of crash. """

    def __init__(self, fn, key="state"):
        super().__init__()
        self.key = key

        self.fn = fn
        makedirs(fn)
        self.canary_fn = fn + ".canary"

        self.db = pickledb.load(fn, auto_dump=False)

        self.state = None

        self._recover()

    def _recover(self):
        if os.path.exists(self.canary_fn):
            self.canary_db = pickledb.load(self.canary_fn, auto_dump=False)
            self.set_state(self.canary_db.get(self.key))
            remove_path(self.canary_fn)

    def get_state(self):
        if self.state is None:
            self.state = self.db.get(self.key) or {}
            return self.state
        return self.state

    def set_state(self, state):
        self.state = state
        self.db.set(self.key, state)
        self.db.dump()

    def __enter__(self):
        self.canary_db = pickledb.load(self.canary_fn, auto_dump=False)
        self.canary_db.set(self.key, self.get_state())
        self.canary_db.dump()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        remove_path(self.canary_fn)


def test_pickle_db_state_provider():
    state_provider = PickleDBStateProvider("state.pickledb")
    state_provider.set_state(16)
    assert state_provider.get_state() == 16

    PickleDBStateProvider("state.pickledb").set_state(17)
    assert PickleDBStateProvider("state.pickledb").get_state() == 17

    with PickleDBStateProvider("state.pickledb") as sp:
        assert sp.get_state() == 17
        sp.set_state(18)
        assert sp.get_state() == 18
    print(state_provider.get_state())

    # clean up
    from utils_ak.os import remove_path

    remove_path("state.pickledb")


if __name__ == "__main__":
    test_pickle_db_state_provider()



from utils_ak.fluid_flow import *


def test_flow_queue_1_in():
    parent = Container("Parent", value=100, max_pressures=[None, 20])

    child1 = Container("Child1", max_pressures=[20, None], limits=[40, None])
    child2 = Container("Child2", max_pressures=[10, None], limits=[50, None])

    queue = Queue("Queue", [child1, child2])

    pipe_connect(parent, queue, "parent-queue")

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_queue_2_out():
    parent1 = Container(
        "Parent1", value=100, max_pressures=[None, 10], limits=[None, 100]
    )
    parent2 = Container(
        "Parent2", value=100, max_pressures=[None, 20], limits=[None, 100]
    )
    queue = Queue("Parent", [parent1, parent2])

    child = Container("Child", max_pressures=[None, None])
    pipe_connect(queue, child, "parent-queue")

    flow = FluidFlow(queue, verbose=True)
    run_flow(flow)


def test_flow_queue_3():
    parent = Container("Parent", value=100, max_pressures=[None, 20])

    child1 = Processor(
        "Child1", max_pressures=[20, None], processing_time=5, limits=[40, None]
    )
    child2 = Processor(
        "Child2", max_pressures=[10, None], processing_time=5, limits=[50, None]
    )

    queue = Queue("Queue", [child1, child2])

    pipe_connect(parent, queue, "parent-queue")

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_queue_4_different_items():
    parent1 = Container(
        "Parent1", item="a", value=100, max_pressures=[None, 10], limits=[None, 100]
    )
    parent2 = Container(
        "Parent2", item="b", value=100, max_pressures=[None, 20], limits=[None, 100]
    )
    queue = Queue("Parent", [parent1, parent2])

    hub = Hub("hub")
    child1 = Container("Child1", item="a", max_pressures=[None, None])
    child2 = Container("Child2", item="b", max_pressures=[None, None])

    pipe_connect(queue, hub, "parent-hub")
    pipe_connect(hub, child1, "hub-child1")
    pipe_connect(hub, child2, "hub-child2")
    flow = FluidFlow(queue, verbose=True)
    run_flow(flow)


if __name__ == "__main__":
    test_flow_queue_1_in()
    # test_flow_queue_2_out()
    # test_flow_queue_3()
    # test_flow_queue_4()



from utils_ak.dag import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.calculations import *


def cast_pipe(pipe_obj):
    if isinstance(pipe_obj, str):
        return Pipe(pipe_obj)
    elif isinstance(pipe_obj, Pipe):
        return pipe_obj
    else:
        raise Exception("Unknown pipe object")


class Pipe(Actor):
    def __init__(self, name=None):
        super().__init__(name)
        self.current_speed = 0
        self.current_item = None
        self.pressures = {"out": None, "in": None}

    @property
    def parent(self):
        if not self.parents:
            return
        assert len(self.parents) == 1
        return self.parents[0]

    @property
    def child(self):
        if not self.children:
            return
        assert len(self.children) == 1
        return self.children[0]

    def update_speed(self, ts):
        self.current_speed = nanmin(self.pressures.values(), require_any=False)

        if np.isnan(self.current_speed):
            self.current_speed = 0

    def __str__(self):
        return f"Pipe {self.name}"

    def stats(self):
        return {"current_speed": self.current_speed, "pressures": self.pressures}

    def set_pressure(self, orient, pressure, item):
        self.pressures[orient] = pressure
        self.current_item = item


class PipeMixin:
    def pipe(self, orient):
        if orient == "in":
            nodes = self.parents
        elif orient == "out":
            nodes = self.children

        if not nodes:
            return
        else:
            assert len(nodes) == 1
            return nodes[0]

    def speed(self, orient):
        if orient == "in":
            if not self.pipe("in"):
                return 0
            return self.pipe("in").current_speed
        elif orient == "out":
            if not self.pipe("out"):
                return 0
            return self.pipe("out").current_speed

    def drain(self):
        return self.speed("in") - self.speed("out")


def pipe_connect(node1, node2, pipe=None):
    if not pipe:
        pipe = cast_pipe(f"{node1} -> {node2}")
    else:
        pipe = cast_pipe(pipe)
    connect(node1, pipe)
    connect(pipe, node2)
    return pipe


def pipe_disconnect(node1, node2):
    # find commmon pipe
    input_pipes = [node for node in node1.children if isinstance(node, Pipe)]
    output_pipes = [node for node in node2.parents if isinstance(node, Pipe)]

    intersection = set(input_pipes) & set(output_pipes)

    assert len(intersection) > 0
    for pipe in intersection:
        disconnect(node1, pipe)
        disconnect(pipe, node2)


def pipe_switch(node1, node2, orient="in"):
    if node1 == node2:
        return
    piped_nodes = [node for node in [node1, node2] if node.pipe(orient)]

    if not piped_nodes:
        return

    pipe1, pipe2 = node1.pipe(orient), node2.pipe(orient)

    if orient == "in":
        disconnect(pipe1, node1)
        disconnect(pipe2, node2)
        connect(pipe1, node2)
        connect(pipe2, node1)
    elif orient == "out":
        disconnect(node1, pipe1)
        disconnect(node2, pipe2)
        connect(node2, pipe1)
        connect(node1, pipe2)



from loguru import logger

from kafka.admin import KafkaAdminClient as VanillaKafkaAdminClient
from kafka.admin import NewTopic
from utils_ak.kafka.kafka_client import KafkaClient

from utils_ak.tqdm import *


class KafkaAdminClient:
    def __init__(self, host):
        self.admin_client = VanillaKafkaAdminClient(bootstrap_servers=host)
        self.client = KafkaClient(host)

    def create_topics(self, topics, num_partitions=1, replication_factor=1):
        new_topics = [
            NewTopic(
                name=topic,
                num_partitions=num_partitions,
                replication_factor=replication_factor,
            )
            for topic in topics
        ]
        return self.admin_client.create_topics(
            new_topics=new_topics, validate_only=False
        )

    def list_topics(self):
        return self.client.consumer.topics()

    def delete_topics(self, topics):
        self.admin_client.delete_topics(topics, timeout_ms=300)
        # todo: check for status


def test():
    from uuid import uuid4

    cli = KafkaAdminClient("localhost:9092")
    new_topic = str(uuid4())
    cli.create_topics([new_topic])
    time.sleep(5)
    assert new_topic in cli.list_topics()
    cli.delete_topics([new_topic])
    time.sleep(1)
    assert new_topic not in cli.list_topics()


def test_print_topics():
    cli = KafkaAdminClient("localhost:9092")
    print(cli.list_topics())


def test_delete_by_pattern(pattern):
    from fnmatch import fnmatch

    cli = KafkaAdminClient("localhost:9092")
    topics = cli.list_topics()
    topics = [topic for topic in topics if fnmatch(topic, pattern)]
    cli.delete_topics(topics)
    new_topics = cli.list_topics()
    assert all([topic not in topics for topic in new_topics])


if __name__ == "__main__":
    # test()
    test_print_topics()
    # test_delete_by_pattern("datasets__*")



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np


# sns.set_style("whitegrid", { 'grid.color': '.9'})


def plot_heatmap(df, ax=None, **kwargs):
    ax = ax or plt.gca()
    sns.heatmap(df, vmax=np.nanmax(df), vmin=np.nanmean(df), **kwargs)
    sns.set(context="paper", font="monospace")
    ax.set_yticklabels(ax.yaxis.get_majorticklabels(), rotation=0)
    ax.set_xticklabels(ax.xaxis.get_majorticklabels(), rotation=90)


def plot_scatter(df, labeled=True, ax=None):
    """ Simple scatter plot. """
    ax = ax or plt.gca()
    df.plot(kind="scatter", x="x", y="y", ax=ax)

    if labeled:
        for k, v in df.iterrows():
            ax.annotate(k, v)



from utils_ak.simple_microservice import SimpleMicroservice

from .worker import Worker


class MicroserviceWorker(Worker):
    """ Working using SimpleMicroservice class from utils_ak.simple_microservice util. """
    def __init__(self, id, payload):
        super().__init__(id, payload)
        self.microservice = SimpleMicroservice(
            id, message_broker=self.payload["message_broker"]
        )

        self.microservice.add_timer(
            self.microservice.publish,
            interval=3.0,
            args=(
                "monitor_in",
                "heartbeat",
            ),
            kwargs={"id": self.id},
        )

        self.microservice.add_timer(
            self.on_start, n_times=1, counter_type="left"
        )  # run once on init

    def send_state(self, status, state):
        self.microservice.logger.debug("Sending state", status=status, state=state)
        self.microservice.publish(
            "monitor_in", "state", id=self.id, status=status, state=state
        )

    async def on_start(self):
        pass

    async def on_input(self, topic, msg):
        raise NotImplementedError

    def run(self):
        self.microservice.run()



import kafka  # todo: move to kafka fully

import uuid

from utils_ak.builtin import update_dic
from copy import deepcopy

from utils_ak.kafka.kafka_binary_search import *


DEFAULT_CONSUMER_CONFIG = {
    "bootstrap_servers": "localhost:9092",
    # "auto_offset_reset": "earliest",
    # "enable.auto.commit": False,
    # "enable.partition.eof": False,
}

DEFAULT_PRODUCER_CONFIG = {
    "bootstrap_servers": "localhost:9092",
    # "queue.buffering.max.ms": 1,
    # "queue.buffering.max.messages": 1000000,
    # "max.in.flight.requests.per.connection": 1,
    # "default.topic.config": {"acks": "all"},
}


class KafkaClient:
    def __init__(self, host=None, consumer_config=None, producer_config=None):
        self.start_offsets = {}

        consumer_config = consumer_config or {}
        self.consumer_config = deepcopy(DEFAULT_CONSUMER_CONFIG)
        self.consumer_config = update_dic(self.consumer_config, consumer_config)
        if host:
            self.consumer_config["bootstrap_servers"] = host
        if "group_id" not in self.consumer_config:
            self.consumer_config["group_id"] = str(uuid.uuid4())
        self.consumer = kafka.KafkaConsumer(**self.consumer_config)

        producer_config = producer_config or {}
        self.producer_config = deepcopy(DEFAULT_PRODUCER_CONFIG)
        self.producer_config = update_dic(self.producer_config, producer_config)
        if host:
            self.producer_config["bootstrap_servers"] = host
        self.producer = kafka.KafkaProducer(**self.producer_config)

        self.init_subscriptions = False

    def subscribe(self, topic, start_offset=None):
        if topic not in self.start_offsets:
            self.start_offsets[topic] = start_offset

    def publish(self, topic, msg):
        self.producer.send(topic, msg)

    def flush(self, timeout=0):
        # self.producer.flush(timeout)
        pass

    def poll(self, timeout=0.0):
        self.start_listening()
        return self.consumer.poll(timeout)

    def start_listening(self):
        if not self.init_subscriptions and self.start_offsets:
            self.consumer.subscribe(list(self.start_offsets.keys()))
            for topic, start_offset in self.start_offsets.items():
                if start_offset is not None:
                    partition = get_single_topic_partition(self.consumer, topic)
                    self.consumer.seek(partition, start_offset)
            self.init_subscriptions = True

    def _get_kafka_topics(self):
        consumer = kafka.KafkaConsumer(
            group_id=str(uuid.uuid4()), bootstrap_servers="localhost:9092"
        )
        return consumer.topics()



from utils_ak.deployment.controller import Controller
from utils_ak.coder import cast_js, cast_dict_or_list
from utils_ak.os import *
from utils_ak.builtin import *


class ProcessController(Controller):
    """ Controller using processes. """

    def __init__(self):
        self.processes = {}

    def start(self, deployment):
        id = deployment["id"]

        assert (
            len(deployment["containers"]) == 1
        ), "Only one-container pods are supported for now"

        entity, container = delistify_single(deployment["containers"].items())
        python_main_path = container["python_main"]
        command_line_arguments = {}
        for k, v in container["command_line_arguments"].items():
            command_line_arguments[k] = v

        cmd = f'python "{python_main_path}"'
        for k, v in command_line_arguments.items():
            cmd += f" --{k} "
            if isinstance(v, str):
                cmd += v
            elif isinstance(v, (dict, list)):
                js = cast_js(v)
                js = js.replace('"', r"\"")
                js = f'"{js}"'
                cmd += js
            elif isinstance(v, bool):
                cmd += "True"
            else:
                raise Exception("Unknown command line argument type")
        self.processes[id] = execute(cmd, is_async=True)

    def stop(self, deployment_id):
        try:
            self.processes[deployment_id].kill()
            self.processes.pop(deployment_id)
        except:
            pass

    def log(self, deployment_id):
        pass



from utils_ak.zmq import ZMQClient, endpoint

client = ZMQClient()


def publisher():
    for i in range(10):
        client.publish(endpoint("*", 5555), "topic", b"msg")


def subscriber():
    client.subscribe(endpoint("localhost", 5555))
    for i in range(10):
        received = None
        while not received:
            received = client.poll()
        print(received)


if __name__ == "__main__":
    import multiprocessing
    import time

    multiprocessing.Process(target=subscriber).start()
    time.sleep(1)
    multiprocessing.Process(target=publisher).start()



import itertools
from utils_ak.message_queue.brokers.broker import Broker


class MultiBroker(Broker):
    def __init__(self, brokers, subscribe_broker, publish_broker):
        self.brokers = brokers
        self.subscribe_broker = subscribe_broker
        self.publish_broker = publish_broker

        self.brokers_iterator = itertools.cycle(self.brokers)

    def publish(self, collection, topic, msg):
        self.publish_broker.publish(collection, topic, msg)

    def subscribe(self, collection, topic):
        self.subscribe_broker.subscribe(collection, topic)

    def poll(self, timeout=0.0):
        cur_broker = next(self.brokers_iterator)
        return cur_broker.poll(timeout)



from utils_ak.kafka.archive.kafka_adapter import KafkaAdapter

import logging

logger = logging.getLogger(__name__)


class KafkaTopicSerialDumper:
    def __init__(
        self,
        topic,
        kafka_adapter,
        table,
        state_provider,
        parser,
        batch=1000,
        serializer=None,
        one_store_limit=None,
    ):
        self.topic = topic
        self.kafka_adapter = kafka_adapter
        self.table = table
        self.state_provider = state_provider
        self.parser = parser
        self.batch = batch
        self.serializer = serializer
        self.state = None
        self.one_store_limit = one_store_limit

    def __str__(self):
        return f"KafkaTopicSerialDumper with topic {self.topic}"

    def store(self):
        self.state = self.state_provider.get_state()
        offset = self.state["offset"]
        self.kafka_adapter.subscribe(self.topic, offset, 0)

        smallest_offset, largest_offset = self.kafka_adapter.get_edge_offsets()
        logger.debug(
            f"smallest_offset={smallest_offset}, largest_offset={largest_offset}"
        )

        if smallest_offset > offset:
            logger.warning(
                f"Missed messages with offsets: {offset} to {smallest_offset}"
            )
            offset = smallest_offset
            self.kafka_adapter.subscribe(self.topic, offset, 0)

        logger.debug(f"Subscribed from offset={offset}")

        total = 0
        last_offset = offset
        finished = False

        sp = self.state_provider

        while not finished:
            kafka_msgs = self.kafka_adapter.receive(self.batch)

            if len(kafka_msgs) == 0:
                finished = True
            else:
                current_offset = None
                for kafka_msg in kafka_msgs:
                    if kafka_msg.offset() > largest_offset or (
                        self.one_store_limit is not None
                        and total >= self.one_store_limit
                    ):
                        finished = True
                        break

                    total += 1

                    if kafka_msg.error() is not None:
                        logging.error(kafka_msg.error())  # TODO: log

                    msg = kafka_msg.value()
                    if self.serializer:
                        msg = self.serializer.decode(msg)

                    for timestamp, key, msg in self.parser(msg, kafka_msg):
                        self.table.store(timestamp, key, msg)

                    current_offset = kafka_msg.offset()
                if current_offset is not None:
                    last_offset = current_offset
                    sp.set_state({"offset": current_offset + 1})

        self.table.flush()
        lag = largest_offset - last_offset
        logger.info(
            f"current_offset={last_offset}, "
            f"largest_offset={largest_offset}, "
            f"lag={lag}"
        )
        return total, lag


if __name__ == "__main__":
    from utils_ak.log import configure_stream_logging

    configure_stream_logging(level=logging.DEBUG)

    from utils_ak.granular_storage.storage import JsonGranularStorage
    from utils_ak.state.provider import PickleDBStateProvider
    from utils_ak.serialization.serializer import JsonSerializer

    sd = KafkaTopicSerialDumper(
        topic="binance_trades",
        kafka_adapter=KafkaAdapter(),
        table=JsonGranularStorage("/tmp/storage")
        .get_db("binance")
        .get_table("binance/user_feed", pattern="%Y%m%d_{base}"),
        state_provider=PickleDBStateProvider(
            "/tmp/binance_trades_state.json", default_state={"offset": 0}
        ),
        parser=lambda value, msg: (value["T"], None, value),
        serializer=JsonSerializer(),
    )
    sd.store()



import portion


def cast_interval(a, b):
    return portion.closedopen(a, b)


def calc_interval_length(interval):
    if interval.empty:
        return 0
    return sum([c.upper - c.lower for c in interval])


if __name__ == "__main__":
    interval = cast_interval(1, 3)
    print(calc_interval_length(interval))



""" Tqdm wrapper with more functionality. """
import time
from datetime import datetime
from tqdm import tqdm as tqdm_vanilla
from tqdm import tqdm_notebook

BACKEND = "tqdm"


try:
    from IPython import get_ipython

    if "IPKernelApp" in get_ipython().config:
        from tqdm import tqdm_notebook as tqdm_backend

        BACKEND = "tqdm_notebook"
except:
    pass


class tqdm:
    def __init__(
        self,
        *args,
        desc=None,
        postfix=None,
        show_timing=False,
        timing_format="%H:%M:%S",
        hook=None,
        **kwargs
    ):
        tqdm_backend = tqdm_notebook if BACKEND == "tqdm_notebook" else tqdm_vanilla
        self.backend = tqdm_backend(*args, **kwargs)
        self.desc = desc
        self.postfix = postfix
        self.hook = hook
        self.show_timing = show_timing
        self.timing_format = timing_format
        self.start = datetime.now()

    def __iter__(self):
        for obj in self.backend.__iter__():
            if self.desc:
                if isinstance(self.desc, str):
                    desc = self.desc
                else:
                    desc = str(self.desc(obj))
            else:
                desc = ""
            postfix_dict = {}
            if self.show_timing:
                timing_str = "[{} - {}] ".format(
                    self.format_dt(self.start), self.format_dt(datetime.now())
                )
                if BACKEND == "tqdm_notebook":
                    postfix_dict["clock"] = timing_str
                else:
                    desc = timing_str + desc
            if desc:
                self.backend.set_description(desc)

            if self.postfix:
                postfix_dict.update(self.postfix(obj))

            if postfix_dict:
                self.backend.set_postfix(postfix_dict)

            if self.hook:
                self.hook(obj)

            yield obj

    def format_dt(self, dt):
        return dt.strftime(self.timing_format)


def test():
    def sample_hook(i):
        if i % 100 == 0:
            print("\nThis is a hook message", i)

    for x in tqdm(
        range(100),
        show_timing=True,
        desc=lambda i: i ** 2,
        postfix=lambda i: {"i": i // 10},
        hook=sample_hook,
    ):
        time.sleep(0.01)


if __name__ == "__main__":
    test()



import re


def def_group(name, pat):
    return r"(?P<{}>{})".format(name, pat)


def ref_group(name):
    return r"(?P={})".format(name)


def replace_with_pattern(s, key, rep_name, rep_pat):
    if key not in s:
        return s
    s = s.replace(key, def_group(rep_name, rep_pat), 1)
    while key in s:
        s = s.replace(key, ref_group(rep_name))
    return s


def non_capturing(pat):
    return r"(?:{})".format(pat)


def or_op(pat1, pat2, capturing=True):
    if capturing:
        return "({}|{})".format(pat1, pat2)
    else:
        return non_capturing("{}|{}".format(pat1, pat2))


def search_many(pattern, s):
    search = re.search(pattern, s)
    if not search:
        return
    return search.groups()


def search_one(pattern, s):
    res = search_many(pattern, s)
    if not res:
        return None
    return res[0]


def test_replace_with_pattern():
    pat = replace_with_pattern("<pat> a <pat> b <pat> c", "<pat>", "foo", ".+")
    print(pat)
    print(re.search(pat, "aaa a aaa b aaa c").groupdict())
    print(re.search(pat, "aaa a aaa b aaa c").groups())
    print(re.search(pat, "asdf a asdf b asdf c").groupdict())
    print(re.search(pat, "asdf1 a asdf2 b asdf3 c") is None)


if __name__ == "__main__":
    test_replace_with_pattern()



from utils_ak.deployment.controller.controller_tests import _test_controller
from utils_ak.deployment.controller.controllers.process_controller import (
    ProcessController,
)


def test_process_controller():
    _test_controller(ProcessController)


if __name__ == "__main__":
    test_process_controller()



import time


def test_broker(broker, collection="collection", topic=""):
    message = "message".encode("utf-8")
    for i in range(100):
        broker.publish(collection, topic, message)

    time.sleep(1)
    broker.subscribe(collection, topic)

    i = 0
    while True:
        messages = broker.poll()

        if not messages:
            continue

        print(i, messages)

        i += 1
        if i == 100:
            break


def test_kafka_broker():
    from utils_ak.message_queue.brokers import KafkaBroker

    test_broker(
        KafkaBroker(consumer_config={"auto_offset_reset": "smallest"}),
        topic="",
    )


def test_multi_broker():
    from utils_ak.message_queue.brokers import KafkaBroker, MultiBroker

    kb1 = KafkaBroker(consumer_config={"auto_offset_reset": "smallest"})
    kb2 = KafkaBroker(consumer_config={"auto_offset_reset": "smallest"})

    mb = MultiBroker([kb1, kb2], kb1, kb2)

    test_broker(mb)


if __name__ == "__main__":
    # test_kafka_broker()
    test_multi_broker()



import logging

from utils_ak.granular_storage.storage.fs_storage import MsgPackGranularStorage
from utils_ak.log import configure_stream_logging

configure_stream_logging(level=logging.INFO)

BROKER = "kafka"


class KafkaTopicLiveDumper(object):
    """
    Live kafka topic backup-er, stores all received messages to a given storage
    """

    def __init__(self, topic, table, timestamp_resolver):
        self.topic = topic
        self.table = table
        self.timestamp_resolver = timestamp_resolver

    def on_msg(self, topic, msg):
        """ Message there is considered to be a dict """

        if self.table is None:
            print(self.topic, msg)
        else:
            timestamp = self.timestamp_resolver(msg)
            self.table.store(timestamp, "", msg)


if __name__ == "__main__":
    topic = "binance_trades"

    dumper = KafkaTopicLiveDumper(
        topic=topic,
        table=MsgPackGranularStorage("/tmp/storage")["binance"].get_table(
            "binance_trades"
        ),
        timestamp_resolver=lambda msg: msg["T"],
    )

    msg = {"value": "hi", "T": 228}

    dumper.on_msg(topic=topic, msg=msg)



from utils_ak.dag import *

from utils_ak.clock import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import *
from utils_ak.fluid_flow.actors.container import Container
from utils_ak.fluid_flow.calculations import *

from functools import wraps


def switch(f):
    @wraps(f)
    def inner(self, *args, **kwargs):
        pipe_switch(self, self.io_containers["in"], "in")
        pipe_switch(self, self.io_containers["out"], "out")

        res = f(self, *args, **kwargs)

        pipe_switch(self, self.io_containers["in"], "in")
        pipe_switch(self, self.io_containers["out"], "out")
        return res

    return inner


class Processor(Actor, PipeMixin):
    def __init__(
        self,
        name,
        items=None,
        processing_time=0,
        transformation_factor=1,
        max_pressures=None,
        limits=None,
    ):
        super().__init__(name)

        max_pressures = max_pressures or [None, None]
        limits = limits or [None, None]
        items = items or ["default", "default"]
        self.io_containers = {
            "in": Container(
                "In",
                item=items[0],
                max_pressures=(max_pressures[0], None),
                limits=(limits[0], None),
            ),
            "out": Container(
                "Out",
                item=items[1],
                max_pressures=(None, max_pressures[1]),
                limits=(None, limits[1]),
            ),
        }
        self._pipe = pipe_connect(
            self.io_containers["in"], self.io_containers["out"], "_pipe"
        )
        self.processing_time = processing_time

        self.last_pipe_speed = None
        self.transformation_factor = transformation_factor

    def is_limit_reached(self, orient):
        return self.io_containers[orient].is_limit_reached(orient)

    def inner_actors(self):
        return [self.io_containers["in"], self._pipe, self.io_containers["out"]]

    def on_set_pressure(self, topic, ts, event):
        self._pipe.set_pressure("out", event["pressure"], event["item"])

    def subscribe(self):
        self.event_manager.subscribe(
            f"processor.set_pressure.{self.id}", self.on_set_pressure
        )

    @switch
    def update_value(self, ts):
        self.io_containers["in"].update_value(ts)
        self.io_containers["out"].update_value(ts, factor=self.transformation_factor)

    @switch
    def update_pressure(self, ts):
        self.io_containers["in"].update_pressure(ts, orients=["in"])
        self.io_containers["out"].update_pressure(ts, orients=["out"])

    @switch
    def update_speed(self, ts):
        self.io_containers["in"].update_speed(ts, set_out_pressure=False)

        if self.processing_time == 0:
            # set new inner pressure at once
            self._pipe.set_pressure(
                "out",
                self.io_containers["in"].speed("in"),
                self.io_containers["out"].item,
            )
        else:
            # set inner pressure delayed with processing time
            if self.last_pipe_speed != self.io_containers["in"].speed("in"):
                self.add_event(
                    f"processor.set_pressure.{self.id}",
                    ts + self.processing_time,
                    {
                        "pressure": self.io_containers["in"].speed("in"),
                        "item": self.io_containers["out"].item,
                    },
                )
                self.last_pipe_speed = self.io_containers["in"].speed("in")

        self._pipe.update_speed(ts)

        self.io_containers["out"].update_speed(ts)

    @switch
    def update_triggers(self, ts):
        for node in [self.io_containers["in"], self.io_containers["out"]]:
            node.update_triggers(ts)

    def __str__(self):
        return f"Processor: {self.name}"

    def stats(self):
        return {
            node.name: node.stats()
            for node in [
                self.io_containers["in"],
                self._pipe,
                self.io_containers["out"],
            ]
        }

    def display_stats(self):
        return [
            self.io_containers["in"].display_stats(),
            self.io_containers["out"].display_stats(),
        ]

    def active_periods(self, orient="in"):
        return self.io_containers[orient].active_periods()



from loguru import logger

from utils_ak.deployment import *

from utils_ak.job_orchestrator.worker.worker_tests import (
    _test_microservice_worker,
    _test_microservice_worker_deployment,
)
from utils_ak.job_orchestrator.worker.sample_worker.sample_worker import *
from utils_ak.job_orchestrator.tests.config.config import config
from utils_ak.job_orchestrator.worker.sample_worker.main import MAIN


def run_batch():
    logger.info("Running batch SampleWorker")
    _test_microservice_worker(
        SampleWorker,
        {"type": "batch", "message_broker": config.TRANSPORT},
        run_listener=True,
    )


def run_streaming():
    logger.info("Running streaming SampleWorker")

    _test_microservice_worker(
        SampleWorker,
        {"type": "streaming", "message_broker": config.TRANSPORT},
        run_listener=True,
    )


def run_deployment():
    logger.info("Running batch SampleWorker using deployment via ProcessController")

    controller = ProcessController()
    _test_microservice_worker_deployment(
        {"type": "batch", "message_broker": config.TRANSPORT},
        MAIN,
        controller,
    )


if __name__ == "__main__":
    run_batch()
    # run_streaming()
    # run_deployment()



import os
from omegaconf import OmegaConf

from utils_ak.job_orchestrator.config import Config


CONFIG_FILES = [
    "configs/.settings.local.yaml",
]

CONFIG_FILES = [os.path.join(os.path.dirname(__file__), fn) for fn in CONFIG_FILES]
omega_confs = [OmegaConf.load(fn) for fn in CONFIG_FILES]
omega_conf = OmegaConf.merge(*omega_confs)
config_dict = OmegaConf.to_container(omega_conf)

config = Config.parse_obj(config_dict)



from datetime import datetime, timedelta
from utils_ak.builtin import update_dic
from utils_ak.time import *
from sortedcollections import SortedList
from utils_ak.numeric import custom_round


class Window:
    def __init__(self):
        self.state = "open"  # 'open', 'closed'

    def close(self):
        self.state = "closed"

    def is_emitable(self):
        raise NotImplementedError

    def is_closeable(self):
        raise NotImplementedError

    def emit(self):
        raise NotImplementedError


class ProcessingSessionWindow(Window):
    def __init__(self, gap):
        """
        :param gap: int, gap between sessions in seconds
        """
        super().__init__()
        self.values = []
        self.last_processing_time = None
        self.gap = gap

    def add(self, value):
        self.values.append(value)
        self.last_processing_time = datetime.now()

    def is_closeable(self):
        return (
            self.last_processing_time
            and (datetime.now() - self.last_processing_time).total_seconds() > self.gap
        )

    def close(self):
        self.state = "closed"
        return self.values


class FieldsCollectorWindow(Window):
    def __init__(self, fields):
        super().__init__()
        self.filled = {}  # {field: {symbol: value}}
        self.fields = fields

    def add(self, values):
        values = {
            k: v for k, v in values.items() if k in self.fields
        }  # remove extra fields
        self.filled = update_dic(self.filled, values)

    def is_closeable(self):
        return set(self.filled.keys()) == set(self.fields)

    def close(self):
        self.state = "closed"
        return self.filled


class CollectorWindow(Window):
    def __init__(
        self,
        window_size,
        timestamp_extractor=lambda msg: msg["value"][0],
    ):
        super().__init__()
        self.timestamp_extractor = timestamp_extractor
        self.buffer = SortedList(key=timestamp_extractor)
        self.window_size = window_size

        self.cur_dt = None
        self.last_dt = None

    def add(self, values):
        if self.cur_dt is None:
            self.cur_dt = round_datetime(
                self.timestamp_extractor(values[0]), self.window_size, "floor"
            )

        for value in values:
            self.buffer.add(value)

        self.last_dt = self.timestamp_extractor(values[-1])

    def is_emitable(self):
        if self.cur_dt is None:
            return False
        return (self.last_dt - self.cur_dt) >= self.window_size

    def emit(self):
        next_ts = self.cur_dt + self.window_size
        index = self.buffer.bisect_left({"value": [next_ts]})
        values = self.buffer[:index]
        del self.buffer[:index]
        res = {"begin": self.cur_dt, "end": next_ts, "period": self.window_size}, values
        self.cur_dt = next_ts
        return res


def test():
    window = FieldsCollectorWindow(fields=["a", "b"])

    print(window.add({"a": 1}))
    print(window.is_closeable(), window.state)
    print(window.add({"b": 1}))
    print(window.is_closeable(), window.state)
    print(window.close())
    print(window.is_closeable(), window.state)

    import time

    window = ProcessingSessionWindow(2)
    window.add("asdf")
    print(window.is_closeable())
    time.sleep(1)
    window.add("asdf")
    print(window.is_closeable())
    time.sleep(1)
    print(window.is_closeable())
    time.sleep(1)
    print(window.is_closeable())
    print(window.close())


def test_collector_window():
    dt = cast_datetime("2020.01.01")
    messages = [
        {
            "source": "binance_timebars_1800",
            "value": [dt + i * timedelta(minutes=1), 1, 2, 3],
        }
        for i in range(10)
    ]

    window = CollectorWindow()
    for message in messages:
        print("Adding message", message)
        window.add([message])
        print(window.buffer)
        print(window.is_emitable())
        if window.is_emitable():
            print(window.emit())
            print(window.buffer)


if __name__ == "__main__":
    # test()

    test_collector_window()



def float_pattern():
    return r"([+-]?[0-9]*[.]?[0-9]+)"


def spaces_on_edge(where="beg"):
    if where == "beg":
        return r"^ *"
    elif where == "end":
        return r" *$"
    else:
        raise Exception("Unknown where")



from .broker import Broker
from .brokers import *


def cast_message_broker(obj):
    if isinstance(obj, Broker):
        return obj
    elif isinstance(obj, (list, tuple)):
        broker_type, broker_config = obj
        return {"zmq": ZMQBroker, "kafka": KafkaBroker}[broker_type](**broker_config)
    else:
        raise Exception("Unknown broker type")



import re


def fill_template(js_obj, **params):
    if isinstance(js_obj, dict):
        return {
            k.format(**params): fill_template(v, **params) for k, v in js_obj.items()
        }
    elif isinstance(js_obj, list):
        return [fill_template(v, **params) for v in js_obj]
    elif isinstance(js_obj, str):
        # check if only replacement
        if re.search(r"^{[^{}]+}$", js_obj):
            return params[js_obj[1:-1]]
        # format as string by default
        return js_obj.format(**params)
    else:
        return js_obj


# todo: tests



import anyconfig
from utils_ak.coder import *


def read_deployment(deployment_fn):
    deployment = anyconfig.load(deployment_fn)
    deployment_dirname = os.path.abspath(os.path.dirname(deployment_fn))
    for container_name, container in deployment["containers"].items():
        if not os.path.exists(container["python_main"]):
            possible_local_fn = os.path.join(
                deployment_dirname, container["python_main"]
            )
            if os.path.exists(possible_local_fn):
                container["python_main"] = possible_local_fn
    return deployment



from utils_ak.job_orchestrator.worker.gen_deployment import create_deployment


def test_gen_deployment():
    print(
        create_deployment(
            "<container_name>",
            "<deployment_id>",
            "<payload>",
            "<image>",
            "<python_main>",
        )
    )


if __name__ == "__main__":
    test_gen_deployment()



from utils_ak.config import cast_config

import os

BASE_DIR = __file__
for i in range(2):
    BASE_DIR = os.path.dirname(BASE_DIR)

CONFIG = None


def load_config(environment):
    global CONFIG
    CONFIG = cast_config(os.path.join(BASE_DIR, "config", f"{environment}.yml"))
    return CONFIG


def test():
    print(CONFIG)
    load_config("test")
    print(CONFIG)


if __name__ == "__main__":
    test()



def cast_prop_values(parent, child, key):
    if not parent:
        pv = None
    else:
        pv = parent[key]

    v = child.relative_props.get(key)
    return pv, v


class MixedProps:
    def __init__(
        self,
        props=None,
        dynamic_accumulators=None,
        dynamic_keys=None,
        static_accumulators=None,
        required_static_keys=None,
    ):

        self.relative_props = props or {}
        self.static_props = {}

        self.dynamic_accumulators = dynamic_accumulators or {}
        # use accumulator keys as default dynamic keys
        self.dynamic_keys = dynamic_keys or list(self.dynamic_accumulators.keys())

        self.static_accumulators = static_accumulators
        self.required_static_keys = required_static_keys or []

        self.parent = None
        self.children = []

    @staticmethod
    def default_accumulator(parent, child, key):
        pv, v = cast_prop_values(parent, child, key)
        return v if v is not None else pv

    def update(self, props, accumulate=False):
        self.relative_props.update(props)
        if accumulate:
            if accumulate is True:
                keys = props.keys()
            elif isinstance(accumulate, list):
                keys = accumulate
            else:
                raise Exception("Unknown accumulation requested")
            self.accumulate_static(keys=keys)

    def accumulate_static(self, keys=None, recursive=False):
        new_keys = keys or []
        new_keys = list(new_keys)

        if not new_keys:
            # reset static props
            self.static_props = {}

        parent_static_props = {} if not self.parent else self.parent.static_props

        all_relevant_keys = (
            list(parent_static_props.keys())
            + list(self.relative_props.keys())
            + self.required_static_keys
        )
        all_relevant_keys = set(all_relevant_keys)
        all_relevant_keys = [
            key for key in all_relevant_keys if key not in self.dynamic_keys
        ]

        # if keys present - filter only necessary keys to accumulate
        new_keys = (
            all_relevant_keys
            if not new_keys
            else [key for key in new_keys if key in all_relevant_keys]
        )

        # update static keys
        for key in new_keys:
            accumulator = self.static_accumulators.get(key, self.default_accumulator)
            self.static_props[key] = accumulator(self.parent, self, key)

        if recursive:
            for child in self.children:
                child.accumulate_static(keys=keys, recursive=recursive)

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def __getitem__(self, item):
        if item in self.dynamic_keys:
            accumulator = self.dynamic_accumulators.get(item, self.default_accumulator)
            return accumulator(self.parent, self, item)
        elif item in self.static_props:
            return self.static_props[item]
        elif item in self.relative_props and item not in self.static_accumulators:
            # item would be in static props if accumulated
            return self.relative_props[item]

    def get(self, item, default=None):
        res = self[item]
        if res is None:
            res = default
        return res


if __name__ == "__main__":

    def t_acc(parent, child, key):
        pv, v = cast_prop_values(parent, child, key)
        pv = pv if pv else 0
        v = v if v else 0
        return pv + v

    def size_acc(parent, child, key):
        size, time_size = child.relative_props.get("size", 0), child.relative_props.get(
            "time_size", 0
        )
        size, time_size = int(size), int(time_size)
        if size:
            return int(size)
        else:
            assert time_size % 5 == 0
            return time_size // 5

    def time_size_acc(parent, child, key):
        size, time_size = child.relative_props.get("size", 0), child.relative_props.get(
            "time_size", 0
        )
        size, time_size = int(size), int(time_size)
        if size:
            return size * 5
        else:
            return time_size

    DYNAMIC_ACCUMULATORS = {"t": t_acc}
    STATIC_ACCUMULATORS = {"size": size_acc, "time_size": time_size_acc}

    def gen_props(props=None):
        return MixedProps(
            props=props,
            dynamic_accumulators=DYNAMIC_ACCUMULATORS,
            static_accumulators=STATIC_ACCUMULATORS,
            required_static_keys=["size", "time_size"],
            dynamic_keys=["t", "b"],
        )

    root = gen_props({"t": 1, "size": 5, "a": "sadf", "d": 1})
    child1 = gen_props({"t": 2})
    child2 = gen_props({"t": 3})
    root.add_child(child1)
    child1.add_child(child2)

    print(root["a"])

    root.accumulate_static(recursive=True)

    print(root.static_props, child1.static_props, root["t"], child1["t"])
    print(root.get("bad key", "bad key default"))

    root.update({"b": 5})
    print(child1["b"])
    print(child2["b"])

    root.update({"c": 6, "d": 7})
    root.accumulate_static(["d"], recursive=True)
    print(child2.static_props)



import copy

from utils_ak.coder import cast_dict_or_list
from utils_ak.numeric import *
from utils_ak.simple_vector import *

from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock
from utils_ak.block_tree.block import Block
from utils_ak.block_tree.validation import validate_disjoint_by_axis

from utils_ak.block_tree.pushers.pushers import *

from loguru import logger


class IterativePusher:
    def __init__(self):
        self.parent = None
        self.block = None

    def init(self):
        pass

    def update(self, results):
        pass

    def __call__(self, *args, **kwargs):
        return self.push(*args, **kwargs)

    def push(self, parent, block, validator, iter_props=None, max_tries=300):
        self.parent = parent
        self.block = block

        iter_props = iter_props or [{}]

        self.init()

        cur_try = 0

        while cur_try < max_tries:
            results = []
            for props in iter_props:
                # try to push
                props = copy.deepcopy(props)

                res = simple_push(parent, block, validator=validator, new_props=props)

                if isinstance(res, Block):
                    # success
                    return block
                else:
                    assert isinstance(res, dict)
                    results.append(res)

            self.update(results)
            cur_try += 1

        logger.info(
            "Failed to push element",
            parent=parent.props["cls"],
            block=block.props["cls"],
        )
        raise AssertionError("Failed to push element")


class AxisPusher(IterativePusher):
    def __init__(self, start_from="last_end", start_shift=0):
        super().__init__()
        self.start_from = start_from
        self.start_shift = start_shift

    def init(self):
        self.axis = self.parent.props["axis"]

        if is_int_like(self.start_from):
            cur_start = int(float(self.start_from))
        elif self.start_from == "beg":
            cur_start = self.block.props["x_rel"][self.axis]
        else:
            if not self.parent.children:
                cur_start = 0
            else:
                if self.start_from == "last_beg":
                    cur_start = max(
                        [
                            child.props["x_rel"][self.axis]
                            for child in self.parent.children
                        ]
                    )
                elif self.start_from == "last_end":
                    cur_start = max(
                        [
                            (child.props["x_rel"] + child.size)[self.axis]
                            for child in self.parent.children
                        ]
                    )
                else:
                    raise Exception("Unknown beg type")
        cur_start += self.start_shift
        cur_start = max(cur_start, 0)

        self.cur_x = cast_simple_vector(self.block.n_dims)
        self.cur_x[self.axis] = cur_start
        self.block.props.update(x=self.cur_x)

    def update(self, results):
        dispositions = [result.get("disposition", None) for result in results]
        dispositions = [d for d in dispositions if d is not None]
        disposition = min(dispositions) if len(dispositions) == len(results) else 1
        # logger.debug("Disposition", disposition=disposition)
        self.cur_x[self.axis] += disposition
        self.block.props.update(x=self.cur_x)


def test_axis_pusher():
    from utils_ak.loguru import configure_loguru_stdout

    configure_loguru_stdout()

    def brute_validator(parent, block):
        for c in parent.children:
            validate_disjoint_by_axis(c, block, axis=parent.props["axis"])

    logger.debug("Dummy push test")
    root = ParallelepipedBlock("root", n_dims=1, x=[2], axis=0)
    a = ParallelepipedBlock("a", n_dims=1, size=[4], axis=0)
    b = ParallelepipedBlock("b", n_dims=1, size=[3], axis=0)

    AxisPusher().push(root, a, brute_validator)
    AxisPusher(start_from=0).push(root, b, brute_validator)
    logger.debug("Root", root=root)


if __name__ == "__main__":
    test_axis_pusher()



import fire
import os
from functools import partial

from utils_ak.loguru import configure_loguru_stdout

from utils_ak.job_orchestrator.worker.sample_worker.sample_worker import SampleWorker
from utils_ak.job_orchestrator.worker.worker import *

MAIN = __file__

if __name__ == "__main__":
    configure_loguru_stdout()
    fire.Fire(partial(run_worker, worker_cls=SampleWorker))



import numpy as np


def is_float(obj):
    return np.issubdtype(type(obj), np.float)


def is_int(obj):
    return np.issubdtype(type(obj), np.integer)


def is_numeric(obj):
    return is_int(obj) or is_float(obj)


def is_int_like(obj):
    if isinstance(obj, str) or is_float(obj):
        try:
            obj = float(obj)
            return int(obj) == obj
        except:
            return False
    elif is_int(obj):
        return True
    else:
        return False


def cast_int(obj, allow_none=False):
    if is_none(obj) and allow_none:
        return obj
    assert is_int_like(obj)

    return int(float(obj))


def test_is_int_like():
    print(is_int_like(1))
    print(is_int_like("1"))
    print(is_int_like("1.0"))
    print(is_int_like("1.2"))
    print(is_int_like("1.2;;"))
    print(cast_int("1.0"))
    print(cast_int(1.0))
    print(cast_int(1))
    print(cast_int(np.int64(1)))


def is_none(obj):
    if obj is None:
        return True
    else:
        try:
            if np.isnan(obj):
                return True
        except:
            pass

    return False


if __name__ == "__main__":
    test_is_int_like()
